# -*- coding: utf-8 -*- #

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os
from functools import partial
import nibabel as nib
from scipy.ndimage import rotate  # to work with NumPy arrays
import numpy as np  # a N-dimensional array object
from skimage.transform import resize
import skimage as sk

# PyQt5 imports
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout
from PyQt5 import QtCore
from PyQt5.QtGui import QImage, QPixmap, QFont
from PyQt5.QtWidgets import (
    QLabel, QScrollArea, QFrame, QSlider, QLineEdit, QSizePolicy, QCheckBox)

# Populse_MIA imports
from populse_mia.utils.tools import ClickableLabel
from populse_mia.software_properties import verCmp
from populse_mia.user_interface.pop_ups import PopUpSelectTag
from populse_mia.software_properties import Config
from populse_mia.user_interface.data_browser import data_browser
from populse_mia.data_manager.project import COLLECTION_CURRENT


class MiniViewer(QWidget):
    """MiniViewer that allows to rapidly visualize scans either with a single
    image per scan with cursors to move in five dimensions or with all images
    of the greater dimension of the scan.

    When the latter is selected, the displayed images depends on their
    dimension:

        - 3D: display all the slices.
        - 4D: display the middle slice of the third dimension for each time
           of the fourth dimension.
        - 5D: display the middle slice of the third dimension for the first
           time of the fourth dimension for each time of the fifth dimension.

    Note:
        - idx corresponds to the index of the displayed image
        - idx in [0, self.max_scans]
        - most of the class's attributes are lists of 0 to self.max_scans
           elements

    Methods:
        - __init__: initialise the MiniViewer object
        - boxSlider: create sliders, their connections and thumbnail labels
          for a selected index
        - changePosValue: change the value of a cursor for the selected index
        - check_box_cursors_state_changed: updates the config file
        - clearLayouts: clear the final layout
        - create_slider: create a slider
        - createDimensionLabels: create the dimension labels for the
          selected index
        - createFieldValue: create a field where will be displayed the
          position of a cursor
        - createLayouts: create the layouts
        - displayPosValue: display the position of each cursor for the
          selected index
        - enableSliders: enable each slider of the selected index
        - image_to_pixmap: create a 2D pixmap from a N-D Nifti image
        - image2DModifications: apply modifications to the image to
          display it correctly
        - indexImage: update the sliders values depending on the size of
          the selected image
        - navigImage: display the 2D image for the selected index
        - openTagsPopUp: opens a pop-up to select the legend of the thumbnails
        - setThumbnail: set the thumbnail tag value under the image frame
        - show_slices: create the thumbnails from the selected file paths
        - update_nb_slices: update the config file and the thumbnails
        - update_visualization_method: update the config file and the
          thumbnails
        - verify_slices: verify the number of selected documents

    """

    def __init__(self, project):
        """Initialise the MiniViewer object

        :param project: current project in the software

        """
        super().__init__()

        self.project = project
        self.first_time = True

        # The MiniViewer is set hidden to give more space to the data_browser
        self.setHidden(True)

        # When multiple selection, limiting the number of thumbnails to
        # max_scans
        self.max_scans = 4

        # Config that allows to read the software preferences
        self.config = Config()

        # Initializing some components of the MiniViewer
        self.labels = QWidget()
        self.frame = QFrame()
        self.scroll_area = QScrollArea()
        self.scroll_area.setWidget(self.frame)
        self.frame_final = QFrame()

        self.label_nb_slices = QLabel()
        self.label_nb_slices.setText("Maximum number of slices: ")

        self.line_edit_nb_slices = QLineEdit()
        self.line_edit_nb_slices.setText(str(self.config.getNbAllSlicesMax()))
        self.line_edit_nb_slices.returnPressed.connect(self.update_nb_slices)

        # All these objects are depending on the number of scans to visualize
        self.im_2D = []
        self.slider_3D = []
        self.slider_4D = []
        self.slider_5D = []
        self.txt_slider_3D = []
        self.txt_slider_4D = []
        self.txt_slider_5D = []
        self.label3D = []
        self.label4D = []
        self.label5D = []
        self.imageLabel = []
        self.img = []
        self.label_description = []

        # Layouts
        self.createLayouts()
        self.setLayout(self.v_box_final)

        # Checkboxes
        self.check_box_slices = QCheckBox('Show all slices (no cursors)')
        if self.config.getShowAllSlices() == True:
            self.check_box_slices.setCheckState(Qt.Checked)
        else:
            self.check_box_slices.setCheckState(Qt.Unchecked)
        self.check_box_slices.stateChanged.connect(
            self.update_visualization_method)

        self.check_box_cursors = QCheckBox('Chain cursors')
        self.check_box_cursors.setToolTip("Allows to connect all cursors "
                                          "when selecting multiple documents")
        if self.config.getChainCursors() == True:
            self.check_box_cursors.setCheckState(Qt.Checked)
        else:
            self.check_box_cursors.setCheckState(Qt.Unchecked)
        self.check_box_cursors.stateChanged.connect(
            self.check_box_cursors_state_changed)

        self.file_paths = ""

    def boxSlider(self, idx):
        """Define horizontal sliders connections and thumbnail labels.

        :param idx: the selected index
        """
        self.slider_3D.insert(idx, self.create_slider(0, 0, 0))
        self.slider_4D.insert(idx, self.create_slider(0, 0, 0))
        self.slider_5D.insert(idx, self.create_slider(0, 0, 0))

        self.slider_3D[idx].valueChanged.connect(
            partial(self.changePosValue, idx, 1))
        self.slider_4D[idx].valueChanged.connect(
            partial(self.changePosValue, idx, 2))
        self.slider_5D[idx].valueChanged.connect(
            partial(self.changePosValue, idx, 3))

        self.txt_slider_3D.insert(idx, self.createFieldValue())
        self.txt_slider_4D.insert(idx, self.createFieldValue())
        self.txt_slider_5D.insert(idx, self.createFieldValue())

    def check_box_cursors_state_changed(self):
        """Update the config file.

        Called when the state of the checkbox to chain the cursors changes.
        """
        if self.check_box_cursors.checkState() == Qt.Checked:
            self.config.setChainCursors(True)
        elif self.check_box_cursors.checkState() == Qt.Unchecked:
            self.config.setChainCursors(False)

    def clearLayouts(self):
        """Clear the final layout"""

        for i in reversed(range(self.v_box_final.count())):
            if self.v_box_final.itemAt(i).widget() is not None:
                self.v_box_final.itemAt(i).widget().setParent(None)

    def changePosValue(self, idx, cursor_to_change):
        """
        Change the value of a cursor for the selected index.

        :param idx: the selected index
        :param cursor_to_change: the cursor to change (1, 2 or 3)
        """
        # If the "Chain cursors" mode is not selected, there is nothing to do
        if self.check_box_cursors.checkState() == Qt.Unchecked:
            self.navigImage(idx)
        else:
            # Checking with cursor has been modified
            if cursor_to_change == 1:
                cursor = self.slider_3D
            elif cursor_to_change == 2:
                cursor = self.slider_4D
            else:
                cursor = self.slider_5D

            # Loop on the thumbnails
            for idx_loop in range(min(self.max_scans, len(self.file_paths))):
                # Disconnecting the connection when changing other cursors
                # values
                cursor[idx_loop].valueChanged.disconnect()
                # Do something only when the cursor is not the one that has
                # been changed by the user
                if idx_loop != idx:
                    if cursor[idx].value() == cursor[idx].maximum():
                        value = cursor[idx_loop].maximum()

                    elif cursor[idx].value() == cursor[idx].minimum():
                        value = cursor[idx_loop].minimum()

                    else:
                        # Updating the new value as the value of the cursor
                        # that has been changed by the user
                        value = round((cursor[idx_loop].maximum() + 1) *
                                      (cursor[idx].value() + 1) /
                                      max(1, cursor[idx].maximum() + 1))
                        value = min(cursor[idx_loop].maximum(), value - 1)
                        value = max(0, int(value))
                    cursor[idx_loop].setValue(value)

                # Changing the image to show
                self.navigImage(idx_loop)
                # Reconnecting
                cursor[idx_loop].valueChanged.connect(
                    partial(self.changePosValue, idx_loop, cursor_to_change))

    def createDimensionLabels(self, idx):
        """Create the dimension labels for the selected index.

        :param idx: the selected index
        """
        font = QFont()
        font.setPointSize(9)

        self.label3D.insert(idx, QLabel())
        self.label4D.insert(idx, QLabel())
        self.label5D.insert(idx, QLabel())

        self.label3D[idx].setFont(font)
        self.label4D[idx].setFont(font)
        self.label5D[idx].setFont(font)

        self.label3D[idx].setText('3D: ')
        self.label4D[idx].setText('4D: ')
        self.label5D[idx].setText('5D: ')

    def createFieldValue(self):
        """Create a field where will be displayed the position of a cursor.

        :return: the corresponding field
        """
        fieldValue = QLineEdit()
        fieldValue.setEnabled(False)
        fieldValue.setFixedWidth(50)
        fieldValue.setAlignment(Qt.AlignCenter)
        fieldValue.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        font = QFont()
        font.setPointSize(9)
        fieldValue.setFont(font)
        return fieldValue

    def createLayouts(self):
        """Create the layouts."""

        self.h_box_images = QHBoxLayout()
        self.h_box_images.setSpacing(10)
        self.v_box = QVBoxLayout()
        self.v_box_final = QVBoxLayout()
        self.h_box_slider_3D = QHBoxLayout()
        self.h_box_slider_4D = QHBoxLayout()
        self.h_box_slider_5D = QHBoxLayout()
        self.v_box_sliders = QVBoxLayout()
        self.h_box = QHBoxLayout()
        self.h_box_check_box = QHBoxLayout()
        self.v_box_thumb = QVBoxLayout()
        self.h_box_thumb = QHBoxLayout()

    def create_slider(self, maxm=0, minm=0, pos=0):
        """Generate an horizontal slider.

        :param maxm: slider's maximum
        :param minm: slider's minimum
        :param pos: slider's initial value
        :return: the slider object
        """
        slider = QSlider(Qt.Horizontal)
        slider.setFocusPolicy(Qt.StrongFocus)
        slider.setTickInterval(1)
        slider.setMaximum(maxm)
        slider.setMinimum(minm)
        slider.setValue(pos)
        slider.setEnabled(False)
        return slider

    def displayPosValue(self, idx):
        """Display the position of each cursor for the selected index.

        :param idx: the selected index
        """
        self.txt_slider_3D[idx].setText(
            str(self.slider_3D[idx].value() + 1) + ' / ' +
            str(self.slider_3D[idx].maximum() + 1))
        self.txt_slider_4D[idx].setText(
            str(self.slider_4D[idx].value() + 1) + ' / ' +
            str(self.slider_4D[idx].maximum() + 1))
        self.txt_slider_5D[idx].setText(
            str(self.slider_5D[idx].value() + 1) + ' / ' +
            str(self.slider_5D[idx].maximum() + 1))

    def enableSliders(self, idx):
        """Enable all the horizontal slider.

        :param idx: the slider's index
        """
        self.slider_3D[idx].setEnabled(True)
        self.slider_4D[idx].setEnabled(True)
        self.slider_5D[idx].setEnabled(True)

    def image_to_pixmap(self, im, i):
        """Create a 2D pixmap from a N-D Nifti image.

        :param im: Nifti image
        :param i: index of the slide
        :return: the corresponding pixmap
        """

        # The image to display depends on the dimension of the image
        # In the 3D case, each slice is displayed
        if len(im.shape) == 3:
            im_2D = im.get_data()[:, :, i].copy()

        # In the 4D case, each middle slice of the 3D dimension is displayed
        # for each time in the 4D dimension
        elif len(im.shape) == 4:
            im_3D = im.get_data()[:, :, :, i].copy()
            middle_slice = int(im_3D.shape[2] / 2)
            im_2D = im_3D[:, :, middle_slice]

        # In the 5D case, each first time of the 4D dimension and
        # its middle slice of the 3D dimension is displayed
        elif len(im.shape) == 5:
            im_4D = im.get_data()[:, :, :, :, i].copy()
            im_3D = im_4D[:, :, :, 1]
            middle_slice = int(im_3D.shape[2] / 2)
            im_2D = im_3D[:, :, middle_slice]

        else:
            im_2D = [0]

        # Making some pixel modification to display the image correctly
        im_2D = self.image2DModifications(0, im_2D)

        w, h = im_2D.shape

        im_Qt = QImage(im_2D.data, w, h, QImage.Format_Indexed8)
        pixm = QPixmap.fromImage(im_Qt)

        return pixm

    def image2DModifications(self, idx, im2D=None):
        """Apply modifications to display the image correctly.

        :param idx: the selected index
        :param im2D: image to modify
        """
        if im2D is not None:
            im2D = rotate(im2D, -90, reshape=False)
            im2D = np.uint8((im2D - im2D.min()) / im2D.ptp() * 255.0)

            # anti_aliasing keyword is defined in skimage since version 0.14.0
            if verCmp(sk.__version__, '0.14.0', 'sup'):
                im2D = resize(im2D, (128, 128), mode='constant',
                              anti_aliasing=False)

            else:
                im2D = resize(im2D, (128, 128), mode='constant')

            im2D = (im2D * 255).astype(np.uint8)
            return im2D

        else:
            self.im_2D[idx] = rotate(self.im_2D[idx], -90, reshape=False)
            self.im_2D[idx] = np.uint8(
                (self.im_2D[idx] - self.im_2D[idx].min())
                / self.im_2D[idx].ptp() * 255.0)

            # anti_aliasing keyword is defined in skimage since version 0.14.0
            if verCmp(sk.__version__, '0.14.0', 'sup'):
                self.im_2D[idx] = resize(self.im_2D[idx], (128, 128),
                                         mode='constant',
                                         anti_aliasing=False)

            else:
                self.im_2D[idx] = resize(self.im_2D[idx], (128, 128),
                                         mode='constant')

            self.im_2D[idx] = (self.im_2D[idx] * 255).astype(np.uint8)

    def indexImage(self, idx):
        """Update all slider values according to the size of the current image.

        :param idx: the selected index
        """
        # Getting the sliders value
        sl3D = self.slider_3D[idx].value()
        sl4D = self.slider_4D[idx].value()
        sl5D = self.slider_5D[idx].value()

        # Depending on the dimension, reading the image data and
        # changing the cursors maximum
        if len(self.img[idx].shape) == 3:
            self.im_2D.insert(
                idx, self.img[idx].get_data()[:, :, sl3D].copy())
            self.slider_3D[idx].setMaximum(self.img[idx].shape[2] - 1)
            self.slider_4D[idx].setMaximum(0)
            self.slider_5D[idx].setMaximum(0)
        if len(self.img[idx].shape) == 4:
            self.im_2D.insert(
                idx, self.img[idx].get_data()[:, :, sl3D, sl4D].copy())
            self.slider_3D[idx].setMaximum(self.img[idx].shape[2] - 1)
            self.slider_4D[idx].setMaximum(self.img[idx].shape[3] - 1)
            self.slider_5D[idx].setMaximum(0)
        if len(self.img[idx].shape) == 5:
            self.im_2D.insert(
                idx, self.img[idx].get_data()[:, :, sl3D, sl4D, sl5D].copy())
            self.slider_3D[idx].setMaximum(self.img[idx].shape[2] - 1)
            self.slider_4D[idx].setMaximum(self.img[idx].shape[3] - 1)
            self.slider_5D[idx].setMaximum(self.img[idx].shape[4] - 1)

    def navigImage(self, idx):
        """Display the 2D image for the selected index.

        :param idx: the selected index
        """
        self.indexImage(idx)
        self.displayPosValue(idx)

        self.image2DModifications(idx)
        w, h = self.im_2D[idx].shape

        image = QImage(self.im_2D[idx].data, w, h, QImage.Format_Indexed8)
        pixm = QPixmap.fromImage(image)
        self.imageLabel[idx].setPixmap(pixm)

    def openTagsPopUp(self):
        """Open a pop-up to select the legend of the thumbnails."""

        self.popUp = PopUpSelectTag(self.project)
        self.popUp.setWindowTitle("Select the image viewer tag")
        if self.popUp.exec_():
            self.verify_slices(self.file_paths)

    def setThumbnail(self, file_path_base_name, idx):
        """Set the thumbnail tag value under the image frame.

        :param file_path_base_name: basename of the selected path
        :param idx: index of the image
        :return:
        """
        # Looking for the tag value to display as a legend of the thumbnail
        for scan in self.project.session.get_documents_names(
                COLLECTION_CURRENT):
            if scan == file_path_base_name:
                value = self.project.session.get_value(
                    COLLECTION_CURRENT, scan, self.config.getThumbnailTag())
                if value is not None:
                    self.label_description[idx].setText(
                        str(value)[:self.nb_char_max])
                else:
                    self.label_description[idx].setText(
                        data_browser.not_defined_value[:self.nb_char_max])
                self.label_description[idx].setToolTip(
                    os.path.basename(self.config.getThumbnailTag()))

    def show_slices(self, file_paths):
        """Creates the thumbnails from the selected file paths.

        :param file_paths: the selected file paths
        """

        # If it's the first time that this function is called, the MiniViewer
        # has to be shown
        if self.first_time:
            self.setHidden(False)
            self.first_time = False

        # If the user has willingly hidden the MiniViewer, the Processes are
        # not made
        if self.isHidden():
            pass
        else:
            self.do_nothing = [False] * len(file_paths)

            self.file_paths = file_paths
            self.max_scans = len(file_paths)

            self.setMinimumHeight(220)

            self.clearLayouts()

            self.frame = QFrame(self)
            self.frame_final = QFrame(self)

            # Limiting the legend of the thumbnails
            self.nb_char_max = 60

            font = QFont()
            font.setPointSize(9)

            # Reading the images from the file paths
            for idx, file_path in enumerate(self.file_paths.copy()):
                try:
                    self.img.insert(idx, nib.load(file_path))
                except nib.filebasedimages.ImageFileError:
                    print("Error while trying to display the image " +
                          file_path)
                    self.file_paths.remove(file_path)
                except FileNotFoundError:
                    print("File " + file_path + " not existing")
                    self.file_paths.remove(file_path)

            # If we are in the "cursors" display mode
            if self.check_box_slices.checkState() == Qt.Unchecked:

                # Layout to aligne each thumbnail (image + cursors)
                self.h_box_thumb = QHBoxLayout()

                # idx represents the index of the selected image
                for idx in range(min(self.max_scans, len(self.file_paths))):
                    if not self.do_nothing[idx]:

                        # Creating sliders and labels
                        self.boxSlider(idx)
                        self.enableSliders(idx)
                        self.createDimensionLabels(idx)

                        # Getting the sliders value and reading the image data
                        self.indexImage(idx)

                        # Making some pixel modification to display the image
                        # correctly
                        self.image2DModifications(idx)

                    self.displayPosValue(idx)

                    w, h = self.im_2D[idx].shape

                    im_Qt = QImage(self.im_2D[idx].data, w, h,
                                   QImage.Format_Indexed8)
                    pixm = QPixmap.fromImage(im_Qt)

                    file_path_base_name = os.path.basename(
                        self.file_paths[idx])

                    # imageLabel is the label where the image is displayed
                    # (as a pixmap)
                    self.imageLabel.insert(idx, QLabel(self))
                    self.imageLabel[idx].setPixmap(pixm)
                    self.imageLabel[idx].setToolTip(file_path_base_name)

                    self.label_description.insert(idx, ClickableLabel())
                    self.label_description[idx].setFont(font)
                    self.label_description[idx].clicked.connect(
                        self.openTagsPopUp)

                    # Looking for the tag value to display as a
                    # legend of the thumbnail
                    file_path_base_name = os.path.relpath(
                        self.file_paths[idx], self.project.folder)
                    self.setThumbnail(file_path_base_name, idx)

                    # Layout that corresponds to the 3D dimension
                    self.h_box_slider_3D = QHBoxLayout()
                    self.h_box_slider_3D.addWidget(self.label3D[idx])
                    self.h_box_slider_3D.addWidget(self.txt_slider_3D[idx])
                    self.h_box_slider_3D.addWidget(self.slider_3D[idx])

                    # Layout that corresponds to the 4D dimension
                    self.h_box_slider_4D = QHBoxLayout()
                    self.h_box_slider_4D.addWidget(self.label4D[idx])
                    self.h_box_slider_4D.addWidget(self.txt_slider_4D[idx])
                    self.h_box_slider_4D.addWidget(self.slider_4D[idx])

                    # Layout that corresponds to the 5D dimension
                    self.h_box_slider_5D = QHBoxLayout()
                    self.h_box_slider_5D.addWidget(self.label5D[idx])
                    self.h_box_slider_5D.addWidget(self.txt_slider_5D[idx])
                    self.h_box_slider_5D.addWidget(self.slider_5D[idx])

                    # Layout for the three sliders
                    self.v_box_sliders = QVBoxLayout()
                    self.v_box_sliders.addLayout(self.h_box_slider_3D)
                    self.v_box_sliders.addLayout(self.h_box_slider_4D)
                    self.v_box_sliders.addLayout(self.h_box_slider_5D)

                    # Layout that corresponds to the image + the sliders
                    self.h_box = QHBoxLayout()
                    self.h_box.addWidget(self.imageLabel[idx])
                    self.h_box.addLayout(self.v_box_sliders)

                    # Layout that corresponds to the image and sliders +
                    # the description
                    self.v_box_thumb = QVBoxLayout()
                    self.v_box_thumb.addLayout(self.h_box)
                    self.v_box_thumb.addWidget(self.label_description[idx])

                    # Layout that will contain all the thumbnails
                    self.h_box_thumb.addLayout(self.v_box_thumb)

                self.frame.setLayout(self.h_box_thumb)

            # If we are in the "all slices" display mode
            else:

                self.h_box_images = QHBoxLayout()
                self.h_box_images.setSpacing(10)
                self.v_box_scans = QVBoxLayout()

                # idx represents the index of the selected image
                for idx in range(len(self.file_paths)):
                    file_path_base_name = os.path.relpath(
                        self.file_paths[idx], self.project.folder)

                    self.label_description.insert(idx, ClickableLabel())
                    self.label_description[idx].setFont(font)
                    self.label_description[idx].clicked.connect(
                        self.openTagsPopUp)

                    # Looking for the tag value to display as a legend
                    # of the thumbnail
                    self.setThumbnail(file_path_base_name, idx)

                    # Depending of the dimension of the image,
                    # the legend of each image and the number of images to
                    # display will change
                    if not self.do_nothing[idx]:
                        if len(self.img[idx].shape) == 3:
                            nb_slices = self.img[idx].shape[2]
                            txt = "Slice n°"
                        elif len(self.img[idx].shape) == 4:
                            nb_slices = self.img[idx].shape[3]
                            txt = "Time n°"
                        elif len(self.img[idx].shape) == 5:
                            nb_slices = self.img[idx].shape[4]
                            txt = "Study n°"
                        else:
                            nb_slices = 0

                        # Limiting the number of images to the number
                        # chosen by the user
                        for i in range(min(nb_slices, int(
                                self.line_edit_nb_slices.text()))):
                            pixm = self.image_to_pixmap(self.img[idx], i)

                            self.v_box = QVBoxLayout()

                            # label corresponds to the label where one image
                            # is displayed
                            label = QLabel(self)
                            label.setPixmap(pixm)
                            label.setToolTip(os.path.basename(
                                self.file_paths[idx]))

                            # Legend of the image (depends on the number
                            # of dimensions)
                            label_info = QLabel()
                            label_info.setFont(font)
                            label_info.setText(txt + str(i + 1))
                            label_info.setAlignment(QtCore.Qt.AlignCenter)

                            self.v_box.addWidget(label)
                            self.v_box.addWidget(label_info)

                            # This layout allows to chain each image
                            self.h_box_images.addLayout(self.v_box)
                        self.v_box_scans.addLayout(self.h_box_images)
                        self.v_box_scans.addWidget(self.label_description[idx])
                self.frame.setLayout(self.v_box_scans)

            # Adding a scroll area if the thumbnails are too large
            self.scroll_area = QScrollArea()
            self.scroll_area.setWidget(self.frame)

            self.h_box_check_box = QHBoxLayout()

            if self.check_box_slices.isChecked():
                self.h_box_check_box.addStretch(1)
                self.label_nb_slices.setHidden(False)
                self.line_edit_nb_slices.setHidden(False)
                self.h_box_check_box.addWidget(self.label_nb_slices)
                self.h_box_check_box.addWidget(self.line_edit_nb_slices)
                self.check_box_cursors.setHidden(True)
            else:
                self.check_box_cursors.setHidden(False)
                self.h_box_check_box.addWidget(self.check_box_cursors)
                self.h_box_check_box.addStretch(1)
                self.label_nb_slices.setHidden(True)
                self.line_edit_nb_slices.setHidden(True)

            self.h_box_check_box.addWidget(self.check_box_slices)

            self.v_box_final.addLayout(self.h_box_check_box)
            self.v_box_final.addWidget(self.scroll_area)

    def update_nb_slices(self):
        """Update the config file and the thumbnails.

        Called when the number of slices to visualize changes.
        """
        nb_slices = self.line_edit_nb_slices.text()
        self.config.setNbAllSlicesMax(nb_slices)
        self.verify_slices(self.file_paths)

    def update_visualization_method(self):
        """Update the config file and the thumbnails.

        Called when the state of the checkbox to show all slices changes.
        """
        if self.check_box_slices.checkState() == Qt.Checked:
            self.config.setShowAllSlices(True)
        elif self.check_box_slices.checkState() == Qt.Unchecked:
            self.config.setShowAllSlices(False)
        self.verify_slices(self.file_paths)

    def verify_slices(self, file_paths):
        """Make 'Show all slices' checkbox unclickable if len(file_paths) > 1.

        :param file_paths: the selected documents
        """
        # Updating the config
        self.config = Config()
        if len(file_paths) > 1:
            self.config.setShowAllSlices(False)
            self.check_box_slices.setCheckState(Qt.Unchecked)
            self.check_box_slices.setCheckable(False)
        else:
            self.check_box_slices.setCheckable(True)
        self.show_slices(file_paths)

    # def check_differences(self, file_paths):
    #     old_to_new = []
    #     self.do_nothing = [False, False, False]
    #     for idx_old, file_path in enumerate(self.file_paths):
    #         if file_path in file_paths:
    #             idx_new = file_paths.index(file_path)
    #             old_to_new.append((idx_old, idx_new))
    #             #if idx_new == idx_old:
    #             self.do_nothing[idx_new] = True
    #
    #     for tp in old_to_new:
    #         idx_old = tp[0]
    #         idx_new = tp[1]
    #         self.shift_thumbnail(idx_old, idx_new)
    #
    # def shift_thumbnail(self, idx_old, idx_new):
    #
    #     self.im_2D[idx_new] = self.im_2D[idx_old]
    #
    #     self.slider_3D[idx_new].setMaximum(self.slider_3D[idx_old].maximum())
    #     self.slider_4D[idx_new].setMaximum(self.slider_4D[idx_old].maximum())
    #     self.slider_5D[idx_new].setMaximum(self.slider_5D[idx_old].maximum())
    #
    #     self.slider_3D[idx_new].setValue(self.slider_3D[idx_old].value())
    #     self.slider_4D[idx_new].setValue(self.slider_4D[idx_old].value())
    #     self.slider_5D[idx_new].setValue(self.slider_5D[idx_old].value())"""
    #
    # def clear_layout(self, layout):
    #     while layout.count() > 0:
    #         item = layout.takeAt(0)
    #         if not item:
    #             continue
    #
    #         w = item.widget()
    #         if w:
    #             w.deleteLater()"""




##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os

# PyQt5 imports
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QTableWidgetItem, QPushButton, QVBoxLayout, \
    QWidget,  QHBoxLayout, QCheckBox, QTableWidget, QComboBox, QLabel

# MIA imports
from populse_mia.user_interface.pipeline_manager.process_mia import ProcessMIA
from populse_mia.user_interface.pop_ups import \
    PopUpSelectTagCountTable
from populse_mia.data_manager.project import COLLECTION_CURRENT, TAG_FILENAME
from populse_mia.utils.tools import ClickableLabel
from populse_mia.software_properties import Config


class IterationTable(QWidget):
    """Widget that handles pipeline iteration.

    Methods:
        - add_tag: adds a tag to visualize in the iteration table
        - emit_iteration_table_updated: emits a signal when the iteration
           scans have been updated
        - fill_values: fill values_list depending on the visualized tags
        - refresh_layout: updates the layout of the widget
        - remove_tag: removes a tag to visualize in the iteration table
        - select_iterated_tag: opens a pop-up to let the user select on which
           tag to iterate
        - select_visualized_tag: opens a pop-up to let the user select which
           tag to visualize in the iteration table
        - update_iterated_tag: updates the widget
        - update_table: updates the iteration table

    """

    iteration_table_updated = pyqtSignal(list)

    def __init__(self, project, scan_list, main_window):
        """
        Initialization of the IterationTable widget

        :param project: current project in the software
        :param scan_list: list of the selected database files
        :param main_window: software's main_window
        """

        QWidget.__init__(self)

        #Necessary for using MIA bricks
        ProcessMIA.project = project

        self.project = project

        if not scan_list:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT)
        else:
            self.scan_list = scan_list

        self.main_window = main_window
        self.iterated_tag = None

        # values_list will contain the different values of each selected tag
        self.values_list = [[], []]

        # Checkbox to choose to iterate the pipeline or not
        self.check_box_iterate = QCheckBox("Iterate pipeline")
        self.check_box_iterate.stateChanged.connect(
            self.emit_iteration_table_updated)

        # Label "Iterate over:"
        self.label_iterate = QLabel("Iterate over:")

        # Label that displays the name of the selected tag
        self.iterated_tag_label = QLabel("Select a tag")

        # Push button to select the tag to iterate
        self.iterated_tag_push_button = QPushButton("Select")
        self.iterated_tag_push_button.clicked.connect(
            self.select_iteration_tag)

        # QComboBox
        self.combo_box = QComboBox()
        self.combo_box.currentIndexChanged.connect(self.update_table)

        # QTableWidget
        self.iteration_table = QTableWidget()

        # Label tag
        self.label_tags = QLabel("Tags to visualize:")

        # Each push button will allow the user to visualize a tag in
        # the iteration browser
        push_button_tag_1 = QPushButton()
        push_button_tag_1.setText("SequenceName")
        push_button_tag_1.clicked.connect(
            lambda: self.select_visualized_tag(0))

        push_button_tag_2 = QPushButton()
        push_button_tag_2.setText("AcquisitionDate")
        push_button_tag_2.clicked.connect(
            lambda: self.select_visualized_tag(1))

        # The list of all the push buttons
        # (the user can add as many as he or she wants)
        self.push_buttons = []
        self.push_buttons.insert(0, push_button_tag_1)
        self.push_buttons.insert(1, push_button_tag_2)

        # Labels to add/remove a tag (a push button)
        self.add_tag_label = ClickableLabel()
        self.add_tag_label.setObjectName('plus')
        sources_images_dir = Config().getSourceImageDir()
        add_tag_picture = QPixmap(os.path.relpath(os.path.join(
            sources_images_dir, "green_plus.png")))
        add_tag_picture = add_tag_picture.scaledToHeight(15)
        self.add_tag_label.setPixmap(add_tag_picture)
        self.add_tag_label.clicked.connect(self.add_tag)

        self.remove_tag_label = ClickableLabel()
        remove_tag_picture = QPixmap(os.path.relpath(os.path.join(
            sources_images_dir, "red_minus.png")))
        remove_tag_picture = remove_tag_picture.scaledToHeight(20)
        self.remove_tag_label.setPixmap(remove_tag_picture)
        self.remove_tag_label.clicked.connect(self.remove_tag)

        # Layout
        self.v_layout = QVBoxLayout()
        self.setLayout(self.v_layout)
        self.refresh_layout()

    def add_tag(self):
        """Add a tag to visualize in the iteration table.

        Used only for tests
        """

        idx = len(self.push_buttons)
        push_button = QPushButton()
        push_button.setText('Tag n°' + str(len(self.push_buttons) + 1))
        push_button.clicked.connect(lambda: self.select_visualized_tag(idx))
        self.push_buttons.insert(len(self.push_buttons), push_button)
        self.refresh_layout()

    def emit_iteration_table_updated(self):
        """
        Emit a signal when the iteration scans have been updated

        """

        if self.check_box_iterate.checkState():
            if hasattr(self, 'scans'):
                self.iteration_table_updated.emit(self.iteration_scans)
            else:
                self.iteration_table_updated.emit(self.scan_list)
        else:
            self.iteration_table_updated.emit(self.scan_list)

    def fill_values(self, idx):
        """
        Fill values_list depending on the visualized tags

        :param idx:
        """

        """ Method that fills the values list when a tag is added
        or removed. """
        tag_name = self.push_buttons[idx].text()
        values = []
        for scan in self.project.session.get_documents_names(
                COLLECTION_CURRENT):
            current_value = self.project.session.get_value(COLLECTION_CURRENT,
                                                           scan, tag_name)
            if current_value is not None:
                values.append(current_value)

        idx_to_fill = len(self.values_list)
        while len(self.values_list) <= idx:
            self.values_list.insert(idx_to_fill, [])
            idx_to_fill += 1

        if self.values_list[idx] is not None:
            self.values_list[idx] = []

        for value in values:
            if value not in self.values_list[idx]:
                self.values_list[idx].append(value)

    def refresh_layout(self):
        """Update the layout of the widget.

        Called in widget's initialization and when a tag push button
        is added or removed.
        """

        first_v_layout = QVBoxLayout()
        first_v_layout.addWidget(self.check_box_iterate)

        second_v_layout = QVBoxLayout()
        second_v_layout.addWidget(self.label_iterate)
        second_v_layout.addWidget(self.iterated_tag_label)

        third_v_layout = QVBoxLayout()
        third_v_layout.addWidget(self.iterated_tag_push_button)
        third_v_layout.addWidget(self.combo_box)

        top_layout = QHBoxLayout()
        top_layout.addLayout(first_v_layout)
        top_layout.addLayout(second_v_layout)
        top_layout.addLayout(third_v_layout)

        self.v_layout.addLayout(top_layout)
        self.v_layout.addWidget(self.iteration_table)

        self.h_box = QHBoxLayout()
        self.h_box.setSpacing(10)
        self.h_box.addWidget(self.label_tags)

        for tag_label in self.push_buttons:
            self.h_box.addWidget(tag_label)

        self.h_box.addWidget(self.add_tag_label)
        self.h_box.addWidget(self.remove_tag_label)
        self.h_box.addStretch(1)

        self.v_layout.addLayout(self.h_box)

    def remove_tag(self):
        """Remove a tag to visualize in the iteration table.

        Used only for tests.
        """
        push_button = self.push_buttons[-1]
        push_button.deleteLater()
        push_button = None
        del self.push_buttons[-1]
        del self.values_list[-1]
        self.refresh_layout()

    def select_iteration_tag(self):
        """
        Open a pop-up to let the user select on which tag to iterate

        """

        ui_select = PopUpSelectTagCountTable(
            self.project,
            self.project.session.get_fields_names(COLLECTION_CURRENT),
            self.iterated_tag)
        if ui_select.exec_():
            self.update_iterated_tag(ui_select.selected_tag)

    def select_visualized_tag(self, idx):
        """Open a pop-up to let the user select which tag to visualize in the
           iteration table

        :param idx: index of the clicked push button
        """

        popUp = PopUpSelectTagCountTable(
            self.project,
            self.project.session.get_fields_names(COLLECTION_CURRENT),
            self.push_buttons[idx].text())
        if popUp.exec_():
            self.push_buttons[idx].setText(popUp.selected_tag)
            self.fill_values(idx)
            self.update_table()

    def update_iterated_tag(self, tag_name):
        """
        Update the widget when the iterated tag is modified

        :param tag_name: name of the iterated tag
        """

        if not self.scan_list:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT)

        self.iterated_tag_push_button.setText(tag_name)
        self.iterated_tag = tag_name
        self.iterated_tag_label.setText(tag_name + ":")

        # Update combo_box
        scans_names = self.project.session.get_documents_names(
            COLLECTION_CURRENT)
        scans_names = list(set(scans_names).intersection(self.scan_list))

        # tag_values_list contains all the values that can take iterated tag
        self.tag_values_list = []
        for scan_name in scans_names:
            tag_value = self.project.session.get_value(COLLECTION_CURRENT,
                                                       scan_name, tag_name)
            if str(tag_value) not in self.tag_values_list:
                self.tag_values_list.append(str(tag_value))

        self.combo_box.clear()
        self.combo_box.addItems(self.tag_values_list)

        self.update_table()

    def update_table(self):
        """
        Update the iteration table

        """

        # Updating the scan list
        if not self.scan_list:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT)

        # Clearing the table and preparing its columns
        self.iteration_table.clear()
        self.iteration_table.setColumnCount(len(self.push_buttons))

        # Headers
        for idx in range(len(self.push_buttons)):
            header_name = self.push_buttons[idx].text()
            if header_name not in self.project.session.get_fields_names(
                    COLLECTION_CURRENT):
                print("{0} not in the project's tags".format(header_name))
                return

            item = QTableWidgetItem()
            item.setText(header_name)
            self.iteration_table.setHorizontalHeaderItem(idx, item)

        # Searching the database scans that correspond to iterated tag value
        filter_query = "({" + self.iterated_tag + "} " + "==" + " \"" + \
                       self.combo_box.currentText() + "\")"
        scans_list = self.project.session.filter_documents(COLLECTION_CURRENT,
                                                           filter_query)
        scans_res = [getattr(document, TAG_FILENAME)
                     for document in scans_list]

        # Taking the intersection between the found database scans and the
        # user selection in the data_browser
        self.iteration_scans = list(
            set(scans_res).intersection(self.scan_list))
        self.iteration_table.setRowCount(len(self.iteration_scans))

        # Filling the table cells
        row = -1
        for scan_name in self.iteration_scans:
            row += 1
            for idx in range(len(self.push_buttons)):
                tag_name = self.push_buttons[idx].text()

                item = QTableWidgetItem()
                item.setText(str(self.project.session.get_value(
                    COLLECTION_CURRENT, scan_name, tag_name)))
                self.iteration_table.setItem(row, idx, item)

        # This will change the scans list in the current Pipeline Manager tab
        self.iteration_table_updated.emit(self.iteration_scans)



# -*- coding: utf-8 -*- #

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

# PyQt5 import
from PyQt5.QtWidgets import QLineEdit

# Populse_MIA imports
from populse_mia.data_manager.project import TAG_FILENAME, TAG_BRICKS


class RapidSearch(QLineEdit):
    """Widget used to search for a pattern in the table (for all the visualized
       tags).

    Enter % to replace any string, _ to replace any character , *Not Defined*
    for the scans with missing value(s).
    Dates are in the following format: yyyy-mm-dd hh:mm:ss.fff”

    Attributes:
        - databrowser: parent data browser widget

    Methods:
        - prepare_filter: prepares the rapid search filter
        - prepare_not_defined_filter: prepares the rapid search filter for not
          defined values
    """

    def __init__(self, databrowser):

        super().__init__()

        self.databrowser = databrowser
        self.setPlaceholderText("Rapid search, enter % to replace any string,"
                                " _ to replace any character, *Not Defined* "
                                "for the scans with missing value(s), "
                                "dates are in the following format: "
                                "yyyy-mm-dd hh:mm:ss.fff")

    def prepare_not_defined_filter(self, tags):
        """Prepare the rapid search filter for not defined values.

        :param tags: list of tags to take into account
        :return: str filter corresponding to the rapid search for not defined
          values
        """

        query = ""

        or_to_write = False

        for tag in tags:

            if tag != TAG_BRICKS:

                if or_to_write:
                    query += " OR "

                query += "({" + tag + "} == null)"

                or_to_write = True

        query += " AND ({" + TAG_FILENAME + "} IN " + str(
            self.databrowser.table_data.scans_to_search).replace("'",
                                                                 "\"") + ")"

        query = "(" + query + ")"

        return query

    @staticmethod
    def prepare_filter(search, tags, scans):
        """Prepare the rapid search filter.

        :param search: Search (str)
        :param tags: List of tags to take into account
        :param scans: List of scans to search into
        :return: str filter corresponding to the rapid search
        """

        query = "("

        or_to_write = False

        for tag in tags:

            if tag != TAG_BRICKS:

                if or_to_write:
                    query += " OR "

                query += "({" + tag + "} LIKE \"%" + search + "%\")"

                or_to_write = True

        query += ") AND ({" + TAG_FILENAME + "} IN " + str(
            scans).replace("'", "\"") + ")"

        return query



# -*- coding: utf-8 -*- #

""" User interface module, encompass the main window and its tabs
definition, as well as pop ups.

Contains:
    Module:
        - main_window
        - pop_ups
        - data_browser
        - pop_ups

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################



# -*- coding: utf-8 -*- #
"""
Module to define pipeline manager tab appearance, settings and methods.

Contains:
    Class:
        - PipelineManagerTab
        - RunProgress
        - RunWorker

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import datetime
import os
import sys
import uuid

from collections import OrderedDict

from PyQt5 import QtCore
from matplotlib.backends.qt_compat import QtWidgets
from traits.trait_errors import TraitError
from traits.api import TraitListObject, Undefined

# PyQt5 imports
from PyQt5.QtCore import Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import (QMenu, QVBoxLayout, QWidget, QSplitter,
                             QApplication, QToolBar, QAction, QHBoxLayout,
                             QScrollArea, QMessageBox, QProgressDialog)

# Capsul imports
from capsul.api import (get_process_instance, StudyConfig, PipelineNode,
                        Switch, NipypeProcess, Pipeline)
from capsul.qt_gui.widgets.pipeline_developper_view import (
    PipelineDevelopperView)

# Populse_MIA imports
from populse_mia.user_interface.pipeline_manager.process_mia import ProcessMIA
from populse_mia.user_interface.pop_ups import PopUpSelectIteration
from populse_mia.user_interface.pipeline_manager.iteration_table import (
    IterationTable)

from populse_mia.data_manager.project import (
    COLLECTION_CURRENT, COLLECTION_INITIAL, COLLECTION_BRICK, BRICK_NAME,
    BRICK_OUTPUTS, BRICK_INPUTS, TAG_BRICKS, BRICK_INIT, BRICK_INIT_TIME,
    TAG_TYPE, TAG_EXP_TYPE, TAG_FILENAME, TAG_CHECKSUM, TYPE_NII, TYPE_MAT)
from populse_mia.user_interface.pipeline_manager.node_controller import (
    NodeController)
from populse_mia.user_interface.pipeline_manager.pipeline_editor import (
    PipelineEditorTabs)
from populse_mia.user_interface.pipeline_manager.process_library import (
    ProcessLibraryWidget)
from populse_mia.software_properties import Config

if sys.version_info[0] >= 3:
    unicode = str

    def values(d):
        return list(d.values())
else:
    def values(d):
        return d.values()


class PipelineManagerTab(QWidget):
    """
    Widget that handles the Pipeline Manager tab.

    Methods:
        - controller_value_changed: update history when a pipeline node is
          changed
        - displayNodeParameters: display the node controller when a node is
          clicked
        - init_pipeline: initialize the current pipeline of the pipeline
          editor
        - layout_view : initialize layout for the pipeline manager
        - loadParameters: load pipeline parameters to the current pipeline of
          the pipeline editor
        - loadPipeline: load a pipeline to the pipeline editor
        - redo: redo the last undone action on the current pipeline editor
        - runPipeline: run the current pipeline of the pipeline editor
        - saveParameters: save the pipeline parameters of the the current
          pipeline of the pipeline editor
        - savePipeline: save the current pipeline of the pipeline editor
        - savePipelineAs: save the current pipeline of the pipeline editor
          under another name
        - undo: undo the last action made on the current pipeline editor
        - updateProcessLibrary: update the library of processes when a
          pipeline is saved
        - update_clinical_mode: update the visibility of widgets/actions
          depending of the chosen mode
        - update_project: update the project attribute of several objects
        - update_scans_list: update the user-selected list of scans
    """

    item_library_clicked = QtCore.Signal(str)

    def __init__(self, project, scan_list, main_window):
        """
        Initialization of the Pipeline Manager tab

        :param project: current project in the software
        :param scan_list: list of the selected database files
        :param main_window: main window of the software
        """

        config = Config()

        # Necessary for using MIA bricks
        ProcessMIA.project = project
        self.project = project
        if not scan_list:
            self.scan_list = self.project.session.get_documents_names(
                COLLECTION_CURRENT)
        else:
            self.scan_list = scan_list
        self.main_window = main_window
        self.enable_progress_bar = False

        # This list is the list of scans contained in the iteration table
        # If it is empty, the scan list in the Pipeline Manager is the scan
        # list from the data_browser
        self.iteration_table_scans_list = []

        QWidget.__init__(self)

        self.verticalLayout = QVBoxLayout(self)
        self.processLibrary = ProcessLibraryWidget(self.main_window)
        self.processLibrary.process_library.item_library_clicked.connect(
            self.item_library_clicked)
        self.item_library_clicked.connect(self._show_preview)

        # self.diagramScene = DiagramScene(self)
        self.pipelineEditorTabs = PipelineEditorTabs(
            self.project, self.scan_list, self.main_window)
        self.pipelineEditorTabs.node_clicked.connect(
            self.displayNodeParameters)
        self.pipelineEditorTabs.switch_clicked.connect(
            self.displayNodeParameters)
        self.pipelineEditorTabs.pipeline_saved.connect(
            self.updateProcessLibrary)
        self.nodeController = NodeController(
            self.project, self.scan_list, self, self.main_window)
        self.nodeController.visibles_tags = \
            self.project.session.get_shown_tags()

        self.iterationTable = IterationTable(
            self.project, self.scan_list, self.main_window)
        self.iterationTable.iteration_table_updated.connect(
            self.update_scans_list)

        self.previewBlock = PipelineDevelopperView(
            pipeline=None, allow_open_controller=False,
            show_sub_pipelines=True, enable_edition=False)

        self.startedConnection = None

        # Actions
        self.load_pipeline_action = QAction("Load pipeline", self)
        self.load_pipeline_action.triggered.connect(self.loadPipeline)

        self.save_pipeline_action = QAction("Save pipeline", self)
        self.save_pipeline_action.triggered.connect(self.savePipeline)

        self.save_pipeline_as_action = QAction("Save pipeline as", self)
        self.save_pipeline_as_action.triggered.connect(self.savePipelineAs)

        self.load_pipeline_parameters_action = QAction(
            "Load pipeline parameters", self)
        self.load_pipeline_parameters_action.triggered.connect(
            self.loadParameters)

        self.save_pipeline_parameters_action = QAction(
            "Save pipeline parameters", self)
        self.save_pipeline_parameters_action.triggered.connect(
            self.saveParameters)

        self.init_pipeline_action = QAction("Initialize pipeline", self)
        self.init_pipeline_action.triggered.connect(self.init_pipeline)

        self.run_pipeline_action = QAction("Run pipeline", self)
        self.run_pipeline_action.triggered.connect(self.runPipeline)

        if config.get_clinical_mode() == True:
            self.save_pipeline_action.setDisabled(True)
            self.save_pipeline_as_action.setDisabled(True)
            self.processLibrary.setHidden(True)
            self.previewBlock.setHidden(True)

        # Initialize toolbar
        self.menu_toolbar = QToolBar()
        self.tags_menu = QMenu()
        self.tags_tool_button = QtWidgets.QToolButton()
        self.scrollArea = QScrollArea()

        # Initialize Qt layout
        self.hLayout = QHBoxLayout()
        self.splitterRight = QSplitter(Qt.Vertical)
        self.splitter0 = QSplitter(Qt.Vertical)
        self.splitter1 = QSplitter(Qt.Horizontal)

        self.layout_view()

        # To undo/redo
        self.nodeController.value_changed.connect(
            self.controller_value_changed)

    def _show_preview(self, name_item):
        self.previewBlock.centerOn(0, 0)
        self.find_process(name_item)

    def add_plug_value_to_database(self, p_value, brick):
        """
        Add the plug value to the database.

        :param p_value: plug value, a file name or a list of file names
        :param brick: brick of the value
        """

        if type(p_value) in [list, TraitListObject]:
            for elt in p_value:
                self.add_plug_value_to_database(elt, brick)
            return

        # This means that the value is not a filename
        if p_value in ["<undefined>", Undefined] or not os.path.isdir(
                os.path.split(p_value)[0]):
            return
        try:
            open(p_value, 'a').close()
        except IOError:
            raise IOError('Could not open {0} file.'.format(p_value))
        else:
            # Deleting the project's folder in the file name so it can
            # fit to the database's syntax
            old_value = p_value
            p_value = p_value.replace(self.project.folder, "")
            p_value = p_value.replace(os.path.abspath(self.project.folder), "")
            if p_value[0] in ["\\", "/"]:
                p_value = p_value[1:]

            # If the file name is already in the database,
            # no exception is raised
            # but the user is warned
            if self.project.session.get_document(COLLECTION_CURRENT, p_value):
                print("Path {0} already in database.".format(p_value))
            else:
                self.project.session.add_document(COLLECTION_CURRENT, p_value)
                self.project.session.add_document(COLLECTION_INITIAL, p_value)

            # Adding the new brick to the output files
            bricks = self.project.session.get_value(
                COLLECTION_CURRENT, p_value, TAG_BRICKS)
            if bricks is None:
                bricks = []
            bricks.append(self.brick_id)
            self.project.session.set_value(
                COLLECTION_CURRENT, p_value, TAG_BRICKS, bricks)
            self.project.session.set_value(
                COLLECTION_INITIAL, p_value, TAG_BRICKS, bricks)

            # Type tag
            filename, file_extension = os.path.splitext(p_value)
            if file_extension == ".nii":
                self.project.session.set_value(
                    COLLECTION_CURRENT, p_value, TAG_TYPE, TYPE_NII)
                self.project.session.set_value(
                    COLLECTION_INITIAL, p_value, TAG_TYPE, TYPE_NII)
            elif file_extension == ".mat":
                self.project.session.set_value(
                    COLLECTION_CURRENT, p_value, TAG_TYPE, TYPE_MAT)
                self.project.session.set_value(
                    COLLECTION_INITIAL, p_value, TAG_TYPE, TYPE_MAT)

            # Adding inherited tags
            if self.inheritance_dict:
                parent_file = self.inheritance_dict[old_value]
                for scan in self.project.session.get_documents_names(
                        COLLECTION_CURRENT):
                    if scan in str(parent_file):
                        database_parent_file = scan
                banished_tags = [TAG_TYPE, TAG_EXP_TYPE, TAG_BRICKS,
                                 TAG_CHECKSUM, TAG_FILENAME]
                for tag in self.project.session.get_fields_names(
                        COLLECTION_CURRENT):
                    if not tag in banished_tags:
                        parent_current_value = self.project.session.get_value(
                            COLLECTION_CURRENT, database_parent_file, tag)
                        self.project.session.set_value(
                            COLLECTION_CURRENT, p_value, tag,
                            parent_current_value)
                        parent_initial_value = self.project.session.get_value(
                            COLLECTION_INITIAL, database_parent_file, tag)
                        self.project.session.set_value(
                            COLLECTION_INITIAL, p_value, tag,
                            parent_initial_value)

            self.project.saveModifications()

    def add_process_to_preview(self, class_process, node_name=None):
        """
        Add a process to the pipeline

        :param class_process: process class's name (str)
        :param node_name: name of the corresponding node
           (using when undo/redo) (str)
        """
        # pipeline = self.previewBlock.scene.pipeline
        pipeline = Pipeline()
        if not node_name:
            class_name = class_process.__name__
            i = 1

            node_name = class_name.lower() + str(i)

            while node_name in pipeline.nodes and i < 100:
                i += 1
                node_name = class_name.lower() + str(i)

            process_to_use = class_process()

        else:
            process_to_use = class_process

        try:
            process = get_process_instance(
                process_to_use)
        except Exception as e:
            print(e)
            return

        pipeline.add_process(node_name, process)
        self.previewBlock.set_pipeline(pipeline)

        # Capsul update
        node = pipeline.nodes[node_name]
        # gnode = self.scene.add_node(node_name, node)

        return node, node_name

    def controller_value_changed(self, signal_list):
        """
        Update history when a pipeline node is changed

        :param signal_list: list of the needed parameters to update history
        """
        case = signal_list.pop(0)

        # For history
        history_maker = []

        if case == "node_name":
            history_maker.append("update_node_name")
            for element in signal_list:
                history_maker.append(element)

        elif case == "plug_value":
            history_maker.append("update_plug_value")
            for element in signal_list:
                history_maker.append(element)

        self.pipelineEditorTabs.undos[
            self.pipelineEditorTabs.get_current_editor()].append(history_maker)
        self.pipelineEditorTabs.redos[
            self.pipelineEditorTabs.get_current_editor()].clear()

        if case == "plug_value":
            node_name = signal_list[0]
            if node_name in ['inputs', 'outputs']:
                node_name = ''

        if case == "node_name":
            node_name = signal_list[1]

        self.nodeController.update_parameters(
            self.pipelineEditorTabs.get_current_pipeline().nodes[
                node_name].process)

    def displayNodeParameters(self, node_name, process):
        """
        Display the node controller when a node is clicked

        :param node_name: name of the node to display parameters
        :param process: process instance of the corresponding node
        :return:
        """
        if isinstance(process, Switch):
            pass
        else:
            self.nodeController.display_parameters(
                node_name, process,
                self.pipelineEditorTabs.get_current_pipeline())
            self.scrollArea.setWidget(self.nodeController)

    def find_process(self, path):
        """
        Find the dropped process in the system's paths

        :param path: class's path (e.g. "nipype.interfaces.spm.Smooth") (str)
        """
        package_name, process_name = os.path.splitext(path)
        process_name = process_name[1:]
        __import__(package_name)
        pkg = sys.modules[package_name]
        for name, instance in sorted(list(pkg.__dict__.items())):
            if name == process_name:
                try:
                    process = get_process_instance(instance)
                except Exception as e:
                    print(e)
                    return
                else:
                    node, node_name = self.add_process_to_preview(instance)
                    gnode = self.previewBlock.scene.gnodes[node_name]
                    gnode.setPos(0, 0)
                    gnode.updateInfoActived(True)
                    # gnode.active = True
                    # gnode.update_node()
                    rect = gnode.sceneBoundingRect()
                    self.previewBlock.scene.setSceneRect(rect)
                    self.previewBlock.fitInView(
                        rect.x(), rect.y(), rect.width() * 1.2,
                                            rect.height() * 1.2,
                        Qt.KeepAspectRatio)
                    self.previewBlock.setAlignment(Qt.AlignCenter)

    def init_pipeline(self, pipeline=None):
        """
        Initialize the current pipeline of the pipeline editor

        :param pipeline: not None if this method is called from a sub-pipeline
        """

        # Should we add a progress bar for the initialization?
        # self.enable_progress_bar = True

        # if self.enable_progress_bar:
        #     self.progress = InitProgress(
        #         self.project, self.pipelineEditorTabs, pipeline)
        #     self.progress.show()
        #     self.progress.exec()
        #
        # else:
        name = os.path.basename(
            self.pipelineEditorTabs.get_current_filename())
        self.main_window.statusBar().showMessage(
            'Pipeline "{0}" is getting initialized. '
            'Please wait.'.format(name))
        QApplication.processEvents()

        config = Config()
        main_pipeline = False

        # If the initialisation is launch for the main pipeline
        if not pipeline:
            pipeline = get_process_instance(
                self.pipelineEditorTabs.get_current_pipeline())
            main_pipeline = True

        # Test, if it works, comment.
        if hasattr(pipeline, 'pipeline_steps'):
            pipeline.pipeline_steps.on_trait_change(
                self.pipelineEditorTabs.get_current_editor()._reset_pipeline,
                remove=True)
        pipeline.on_trait_change(
            self.pipelineEditorTabs.get_current_editor()._reset_pipeline,
            'selection_changed', remove=True)
        pipeline.on_trait_change(
            self.pipelineEditorTabs.get_current_editor()._reset_pipeline,
            'user_traits_changed', remove=True)

        # nodes_to_check contains the node names that need to be update
        nodes_to_check = []

        # nodes_inputs_ratio is a dictionary whose keys are the node names
        # and whose values are a list of two elements: the first one being
        # the number of activated mandatory input plugs, the second one being
        # the total number of mandatory input plugs of the corresponding node
        nodes_inputs_ratio = {}

        # nodes_inputs_ratio_list contains the ratio between the number of
        # activated mandatory input plugs and the total number of mandatory
        # input plugs of the corresponding node (the order is the same as
        # nodes_to_check)
        nodes_inputs_ratio_list = []

        for node_name, node in pipeline.nodes.items():
            # Updating the project attribute of the processes
            if hasattr(node, 'process'):
                process = node.process
                if hasattr(process, 'use_project') and process.use_project:
                    process.project = self.project

            nb_plugs_from_in = 0
            nb_plugs = 0

            for plug_name, plug in node.plugs.items():
                if plug.links_from and not plug.output and not plug.optional:
                    nb_plugs += 1
                    # If the link come from the pipeline "global" inputs, it is
                    # added to compute the ratio
                    if list(plug.links_from)[0][0] == "":
                        nb_plugs_from_in += 1

            if nb_plugs == 0:
                ratio = 0
            else:
                ratio = nb_plugs_from_in / nb_plugs

            nodes_to_check.append(node_name)
            nodes_inputs_ratio[node_name] = [nb_plugs_from_in, nb_plugs]
            nodes_inputs_ratio_list.append(ratio)

        # Sorting the nodes_to_check list as the order (the nodes having
        # the highest ratio
        # being at the end of the list)
        nodes_to_check = [x for _, x in sorted(
            zip(nodes_inputs_ratio_list, nodes_to_check))]

        while nodes_to_check:
            # Finding one node that has a ratio of 1,
            # which means that all of its mandatory
            # inputs are "connected"
            key_name = [key for key, value in nodes_inputs_ratio.items() if
                        value[0] == value[1]]
            if key_name:
                # This node can be initialized so it is placed at the
                # end of the nodes_to_check list
                nodes_to_check.append(key_name[0])

                # It can also be removed from the dictionary
                del nodes_inputs_ratio[key_name[0]]

            # Reversing the list so that the node to be initialized is
            # at the first place
            # Using OrderedDict allows to remove the duplicate in the list
            # without losing the order. So if key_name[0] appears twice,
            # it will stay at the first place
            nodes_to_check = list(
                OrderedDict(
                    (x, True) for x in nodes_to_check[::-1]).keys())

            node_name = nodes_to_check.pop(0)

            nodes_to_check = nodes_to_check[::-1]

            # Inputs/Outputs nodes will be automatically updated with
            # the method update_nodes_and_plugs_activation of the
            # pipeline object
            if node_name in ['', 'inputs', 'outputs']:
                continue

            # If the node is a pipeline node,
            # each of its nodes has to be initialised
            node = pipeline.nodes[node_name]
            if isinstance(node, PipelineNode):
                sub_pipeline = node.process
                self.init_pipeline(sub_pipeline)

                for plug_name in node.plugs.keys():

                    # If the plug is an output and is
                    if hasattr(node.plugs[plug_name], 'links_to'):
                        # connected to another one
                        list_info_link = list(
                            node.plugs[plug_name].links_to)

                        for info_link in list_info_link:

                            # The third element of info_link contains the
                            if info_link[2] in pipeline.nodes.values():
                                # destination node object
                                dest_node_name = info_link[0]

                                if dest_node_name:
                                    # Adding the destination node name and
                                    # incrementing the input counter of the
                                    # latter if it is not the pipeline
                                    # "global" outputs ('')
                                    nodes_to_check.append(dest_node_name)
                                    nodes_inputs_ratio[dest_node_name][
                                        0] += 1

                pipeline.update_nodes_and_plugs_activation()
                continue

            # Adding the brick to the bricks history
            self.brick_id = str(uuid.uuid4())
            self.project.session.add_document(COLLECTION_BRICK,
                                              self.brick_id)
            self.project.session.set_value(
                COLLECTION_BRICK, self.brick_id, BRICK_NAME, node_name)
            self.project.session.set_value(
                COLLECTION_BRICK, self.brick_id, BRICK_INIT_TIME,
                datetime.datetime.now())
            self.project.session.set_value(
                COLLECTION_BRICK, self.brick_id, BRICK_INIT, "Not Done")
            self.project.saveModifications()

            process = node.process
            process_name = str(process).split('.')[0].split('_')[0][1:]

            print('\nUpdating the launching parameters for {0} '
                  'process node: {1} ...\n'.format(process_name,
                                                   node_name))

            # TODO 'except' instead of 'if' to test matlab launch ?
            # Test for matlab launch
            if config.get_use_spm_standalone() == True:
                pipeline.nodes[node_name].process.use_mcr = True
                pipeline.nodes[node_name].process.paths = \
                    config.get_spm_standalone_path().split()
                pipeline.nodes[node_name].process.matlab_cmd = \
                    config.get_matlab_command()

            elif config.get_use_spm() == True:
                pipeline.nodes[node_name].process.use_mcr = False
                pipeline.nodes[node_name].process.paths = \
                    config.get_spm_path().split()
                pipeline.nodes[node_name].process.matlab_cmd = \
                    config.get_matlab_command()

            # Test for matlab launch
            if not os.path.isdir(
                    os.path.abspath(
                        self.project.folder + os.sep + 'scripts')):
                os.mkdir(os.path.abspath(
                    self.project.folder + os.sep + 'scripts'))

            pipeline.nodes[node_name].process.output_directory = \
                os.path.abspath(
                    self.project.folder + os.sep + 'scripts')
            pipeline.nodes[node_name].process.mfile = True

            # Getting the list of the outputs of the node according
            # to its inputs
            try:
                self.inheritance_dict = None
                (process_outputs,
                 self.inheritance_dict) = process.list_outputs()
            except TraitError as error:
                print("TRAIT ERROR for node {0}".format(node_name))
                print(error)

            except ValueError:
                process_outputs = process.list_outputs()
                print("No inheritance dict for the "
                      "process {0}.".format(node_name))

            # If the process has no "list_outputs" method,
            # which is the case for Nipype's
            except AttributeError:
                # interfaces
                try:
                    process_outputs = process._nipype_interface._list_outputs()
                    # The Nipype Process outputs are always
                    # "private" for Capsul
                    tmp_dict = {}
                    for key, value in process_outputs.items():
                        tmp_dict['_' + key] = process_outputs[key]
                    process_outputs = tmp_dict
                # TODO: test which kind of error can generate Nipype interface
                except Exception:
                    print("No output list method for "
                          "the process {0}.".format(node_name))
                    process_outputs = {}

            # Adding I/O to database history
            inputs = process.get_inputs()

            for key in inputs:

                if inputs[key] is Undefined:
                    inputs[key] = "<undefined>"

                # Could be cleaner to do some tests with warning pop up if
                # necessary (ex. if config.get_spm_standalone_path() is
                # None, etc ...)

                # if (isinstance(process, NipypeProcess)) and (key in
                # keys2consider) and (inputs[key] == "<undefined>"):

                # update launching parameters for IRMaGe_processes bricks
                # Test for matlab launch
                if 'NipypeProcess' in str(process.__class__):
                    print('\nUpdating the launching parameters for nipype '
                          'process node: {0} ...'.format(node_name))
                    # plugs to be filled automatically
                    keys2consider = ['use_mcr', 'paths',
                                     'matlab_cmd', 'output_directory']
                    # use_mcr parameter
                    if (key == keys2consider[0]) and (
                            config.get_use_spm_standalone() == True):
                        inputs[key] = True
                    elif (key == keys2consider[0]) and (
                            config.get_use_spm_standalone() == False):
                        inputs[key] = False
                    # paths  parameter
                    if (key == keys2consider[1]) and (
                            config.get_use_spm_standalone() == True):
                        inputs[
                            key] = config.get_spm_standalone_path().split()
                    elif (key == keys2consider[1]) and (
                            config.get_use_spm() == True):
                        inputs[key] = config.get_spm_path().split()
                    # matlab_cmd parameter
                    if (key == keys2consider[2]) and (
                            config.get_use_spm_standalone() == True):
                        inputs[key] = (config.get_spm_standalone_path() +
                                       '/run_spm12.sh ' +
                                       config.get_matlab_standalone_path() +
                                       ' script')
                    elif key == keys2consider[2] and \
                            config.get_use_spm_standalone() == False:
                        inputs[key] = config.get_matlab_path()

                    # output_directory parameter
                    if key == keys2consider[3]:

                        if not os.path.isdir(os.path.abspath(
                                self.project.folder + '/scripts')):
                            os.mkdir(os.path.abspath(
                                self.project.folder + '/scripts'))

                        inputs[key] = os.path.abspath(
                            self.project.folder + '/scripts')

                    try:
                        pipeline.nodes[node_name].set_plug_value(
                            key, inputs[key])

                    except TraitError:

                        if isinstance(inputs[key], list) and len(
                                inputs[key]) == 1:

                            try:
                                pipeline.nodes[key].set_plug_value(
                                    key, inputs[key][0])

                            except TraitError:
                                print("Trait error for {0} plug of {1} "
                                      "node".format(key, inputs[key]))

            outputs = process.get_outputs()
            for key in outputs:
                value = outputs[key]
                if value is Undefined:
                    outputs[key] = "<undefined>"
            self.project.saveModifications()
            self.project.session.set_value(COLLECTION_BRICK, self.brick_id,
                                           BRICK_INPUTS, inputs)
            self.project.session.set_value(COLLECTION_BRICK, self.brick_id,
                                           BRICK_OUTPUTS, outputs)

            if process_outputs:
                for plug_name, plug_value in process_outputs.items():
                    node = pipeline.nodes[node_name]
                    if plug_name not in node.plugs.keys():
                        continue
                    if plug_value not in ["<undefined>", Undefined]:
                        self.add_plug_value_to_database(plug_value,
                                                        self.brick_id)

                    list_info_link = list(node.plugs[plug_name].links_to)

                    # If the output is connected to another node,
                    # the latter is added to nodes_to_check
                    for info_link in list_info_link:
                        dest_node_name = info_link[0]
                        if dest_node_name:
                            # Adding the destination node name and incrementing
                            # the input counter of the latter
                            nodes_to_check.append(dest_node_name)
                            nodes_inputs_ratio[dest_node_name][0] += 1

                    try:
                        pipeline.nodes[node_name].set_plug_value(plug_name,
                                                                 plug_value)
                    except TraitError:
                        if type(plug_value) is list and len(
                                plug_value) == 1:
                            try:
                                pipeline.nodes[node_name].set_plug_value(
                                    plug_name, plug_value[0])
                            except TraitError:
                                print("Trait error for {0} plug of {1} "
                                      "node".format(plug_name, node_name))
                                pass

                    pipeline.update_nodes_and_plugs_activation()

            # Adding I/O to database history again to update outputs
            inputs = process.get_inputs()
            for key in inputs:
                value = inputs[key]
                if value is Undefined:
                    inputs[key] = "<undefined>"
            outputs = process.get_outputs()
            for key in outputs:
                value = outputs[key]
                if value is Undefined:
                    outputs[key] = "<undefined>"
            self.project.session.set_value(COLLECTION_BRICK, self.brick_id,
                                           BRICK_INPUTS, inputs)
            self.project.session.set_value(COLLECTION_BRICK, self.brick_id,
                                           BRICK_OUTPUTS, outputs)

            # Setting brick init state if init finished correctly
            self.project.session.set_value(COLLECTION_BRICK, self.brick_id,
                                           BRICK_INIT, "Done")
            self.project.saveModifications()

        # Test, if it works, comment.
        pipeline.on_trait_change(
            self.pipelineEditorTabs.get_current_editor()._reset_pipeline,
            'selection_changed', dispatch='ui')
        pipeline.on_trait_change(
            self.pipelineEditorTabs.get_current_editor()._reset_pipeline,
            'user_traits_changed', dispatch='ui')
        if hasattr(pipeline, 'pipeline_steps'):
            pipeline.pipeline_steps.on_trait_change(
                self.pipelineEditorTabs.get_current_editor()._reset_pipeline,
                dispatch='ui')

        # Updating the node controller
        if main_pipeline:
            node_controller_node_name = self.nodeController.node_name
            if node_controller_node_name in ['inputs', 'outputs']:
                node_controller_node_name = ''
            self.nodeController.display_parameters(
                self.nodeController.node_name,
                pipeline.nodes[node_controller_node_name].process,
                pipeline)
        self.main_window.statusBar().showMessage(
            'Pipeline "{0}" has been initialized.'.format(name))

    def layout_view(self):
        """Initialize layout for the pipeline manager tab"""
        self.setWindowTitle("Diagram editor")

        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setWidget(self.nodeController)

        # Toolbar
        self.tags_menu.addAction(self.load_pipeline_action)
        self.tags_menu.addAction(self.save_pipeline_action)
        self.tags_menu.addAction(self.save_pipeline_as_action)
        self.tags_menu.addSeparator()
        self.tags_menu.addAction(self.load_pipeline_parameters_action)
        self.tags_menu.addAction(self.save_pipeline_parameters_action)
        self.tags_menu.addSeparator()
        self.tags_menu.addAction(self.init_pipeline_action)
        self.tags_menu.addAction(self.run_pipeline_action)

        self.tags_tool_button.setText('Pipeline')
        self.tags_tool_button.setPopupMode(
            QtWidgets.QToolButton.MenuButtonPopup)
        self.tags_tool_button.setMenu(self.tags_menu)

        # Layouts

        self.hLayout.addWidget(self.tags_tool_button)
        self.hLayout.addStretch(1)

        self.splitterRight.addWidget(self.iterationTable)
        self.splitterRight.addWidget(self.scrollArea)
        self.splitterRight.setSizes([400, 400])

        # previewScene = QGraphicsScene()
        # previewScene.setSceneRect(QtCore.QRectF())
        # self.previewDiagram = QGraphicsView()
        # self.previewDiagram.setEnabled(False)

        self.splitter0.addWidget(self.processLibrary)
        self.splitter0.addWidget(self.previewBlock)

        self.splitter1.addWidget(self.splitter0)
        self.splitter1.addWidget(self.pipelineEditorTabs)
        self.splitter1.addWidget(self.splitterRight)
        self.splitter1.setSizes([200, 800, 200])

        # self.splitter2 = QSplitter(Qt.Vertical)
        # self.splitter2.addWidget(self.splitter1)
        # self.splitter2.setSizes([800, 100])

        self.verticalLayout.addLayout(self.hLayout)
        self.verticalLayout.addWidget(self.splitter1)

    def loadPipeline(self):
        """
        Load a pipeline to the pipeline editor

        """
        self.pipelineEditorTabs.load_pipeline()

    def loadParameters(self):
        """
        Load pipeline parameters to the current pipeline of the pipeline editor

        """
        self.pipelineEditorTabs.load_pipeline_parameters()

        self.nodeController.update_parameters()

    def redo(self):
        """
        Redo the last undone action on the current pipeline editor

        Actions that can be redone:
            - add_process
            - delete_process
            - export_plug
            - export_plugs
            - remove_plug
            - update_node_name
            - update_plug_value
            - add_link
            - delete_link

        """
        c_e = self.pipelineEditorTabs.get_current_editor()

        # We can redo if we have an action to make again
        if len(self.pipelineEditorTabs.redos[c_e]) > 0:
            to_redo = self.pipelineEditorTabs.redos[c_e].pop()
            # The first element of the list is the type of action made
            # by the user
            action = to_redo[0]

            if action == "delete_process":
                node_name = to_redo[1]
                class_process = to_redo[2]
                links = to_redo[3]
                c_e.add_process(
                    class_process, node_name, from_redo=True, links=links)

            elif action == "add_process":
                node_name = to_redo[1]
                c_e.del_node(node_name, from_redo=True)

            elif action == "export_plug":
                temp_plug_name = to_redo[1]
                c_e._remove_plug(
                    _temp_plug_name=temp_plug_name, from_redo=True)

            elif action == "export_plugs":
                # No redo possible
                pass

            elif action == "remove_plug":
                temp_plug_name = to_redo[1]
                new_temp_plugs = to_redo[2]
                optional = to_redo[3]
                c_e._export_plug(temp_plug_name=new_temp_plugs[0],
                                 weak_link=False,
                                 optional=optional,
                                 from_redo=True,
                                 pipeline_parameter=temp_plug_name[1])

                # Connecting all the plugs that were connected
                # to the original plugs
                for plug_tuple in new_temp_plugs:
                    # Checking if the original plug is a pipeline
                    # input or output to adapt
                    # the links to add.
                    if temp_plug_name[0] == 'inputs':
                        source = ('', temp_plug_name[1])
                        dest = plug_tuple
                    else:
                        source = plug_tuple
                        dest = ('', temp_plug_name[1])

                    c_e.scene.add_link(source, dest, active=True, weak=False)

                    # Writing a string to represent the link
                    source_parameters = ".".join(source)
                    dest_parameters = ".".join(dest)
                    link = "->".join((source_parameters, dest_parameters))

                    c_e.scene.pipeline.add_link(link)
                    c_e.scene.update_pipeline()

            elif action == "update_node_name":
                node = to_redo[1]
                new_node_name = to_redo[2]
                old_node_name = to_redo[3]
                c_e.update_node_name(
                    node, new_node_name, old_node_name, from_redo=True)

            elif action == "update_plug_value":
                node_name = to_redo[1]
                new_value = to_redo[2]
                plug_name = to_redo[3]
                value_type = to_redo[4]
                c_e.update_plug_value(
                    node_name, new_value, plug_name,
                    value_type, from_redo=True)

            elif action == "add_link":
                link = to_redo[1]
                c_e._del_link(link, from_redo=True)

            elif action == "delete_link":
                source = to_redo[1]
                dest = to_redo[2]
                active = to_redo[3]
                weak = to_redo[4]
                c_e.add_link(source, dest, active, weak, from_redo=True)
                # link = source[0] + "." + source[1]
                # + "->" + dest[0] + "." + dest[1]

            c_e.scene.pipeline.update_nodes_and_plugs_activation()
            self.nodeController.update_parameters()

    def runPipeline(self):
        """
        Run the current pipeline of the pipeline editor

        """
        name = os.path.basename(self.pipelineEditorTabs.get_current_filename())
        self.main_window.statusBar().showMessage(
            'Pipeline "{0}" is getting run. Please wait.'.format(name))
        QApplication.processEvents()

        if self.iterationTable.check_box_iterate.isChecked():
            iterated_tag = self.iterationTable.iterated_tag
            tag_values = self.iterationTable.tag_values_list
            ui_iteration = PopUpSelectIteration(iterated_tag, tag_values)
            if ui_iteration.exec():
                tag_values = ui_iteration.final_values
                pipeline_progress = dict()
                pipeline_progress['size'] = len(tag_values)
                pipeline_progress['counter'] = 1
                pipeline_progress['tag'] = iterated_tag
                for tag_value in tag_values:
                    # Status bar update
                    pipeline_progress['tag_value'] = tag_value

                    idx_combo_box = self.iterationTable.combo_box.findText(
                        tag_value)
                    self.iterationTable.combo_box.setCurrentIndex(
                        idx_combo_box)
                    self.iterationTable.update_table()

                    self.init_pipeline()
                    self.main_window.statusBar().showMessage(
                        'Pipeline "{0}" is getting run for {1} {2}. '
                        'Please wait.'.format(name, iterated_tag, tag_value))
                    QApplication.processEvents()
                    self.progress = RunProgress(self.pipelineEditorTabs,
                                                pipeline_progress)
                    self.progress.show()
                    self.progress.exec()
                    pipeline_progress['counter'] += 1
                    # # self.init_pipeline(self.pipeline)
                    # idx = self.progress.value()
                    # idx += 1
                    # self.progress.setValue(idx)
                    # QApplication.processEvents()

            self.main_window.statusBar().showMessage(
                'Pipeline "{0}" has been run for {1} {2}. Please wait.'.format(
                    name, iterated_tag, tag_values))
        else:
            try:
                self.progress = RunProgress(self.pipelineEditorTabs)
                self.progress.show()
                self.progress.exec()
                
            except Exception as e:
                print('\n When the pipeline was launched, the following '
                      'exception was raised: {0} ...'.format(e, ))
                self.main_window.statusBar().showMessage(
                    'Pipeline "{0}" has not been correctly run.'.format(name))
                
            else:
                self.main_window.statusBar().showMessage(
                    'Pipeline "{0}" has been correctly run.'.format(name))

    def saveParameters(self):
        """
        Save the pipeline parameters of the the current pipeline of the
        pipeline editor

        """
        self.pipelineEditorTabs.save_pipeline_parameters()

    def savePipeline(self):
        """
        Save the current pipeline of the pipeline editor

        """
        self.main_window.statusBar().showMessage(
            'The pipeline is getting saved. Please wait.')
        # QApplication.processEvents()

        filename = self.pipelineEditorTabs.get_current_filename()

        if filename:
            self.pipelineEditorTabs.save_pipeline(new_file_name=filename)

        else:
            self.pipelineEditorTabs.save_pipeline()

        self.main_window.statusBar().showMessage(
            'The pipeline has been saved.')

    def savePipelineAs(self):
        """
        Save the current pipeline of the pipeline editor under another name

        """

        self.main_window.statusBar().showMessage(
            'The pipeline is getting saved. Please wait.')

        self.pipelineEditorTabs.save_pipeline()

        self.main_window.statusBar().showMessage(
            'The pipeline has been saved.')

    def undo(self):
        """
        Undo the last action made on the current pipeline editor

        Actions that can be undone:
            - add_process
            - delete_process
            - export_plug
            - export_plugs
            - remove_plug
            - update_node_name
            - update_plug_value
            - add_link
            - delete_link

        """
        c_e = self.pipelineEditorTabs.get_current_editor()

        # We can undo if we have an action to revert
        if len(self.pipelineEditorTabs.undos[c_e]) > 0:
            to_undo = self.pipelineEditorTabs.undos[c_e].pop()
            # The first element of the list is the type of action made
            # by the user
            action = to_undo[0]

            if action == "add_process":
                node_name = to_undo[1]
                c_e.del_node(node_name, from_undo=True)

            elif action == "delete_process":
                node_name = to_undo[1]
                class_name = to_undo[2]
                links = to_undo[3]
                c_e.add_process(
                    class_name, node_name, from_undo=True, links=links)

            elif action == "export_plug":
                temp_plug_name = to_undo[1]
                c_e._remove_plug(
                    _temp_plug_name=temp_plug_name, from_undo=True)

            elif action == "export_plugs":
                parameter_list = to_undo[1]
                node_name = to_undo[2]
                for parameter in parameter_list:
                    temp_plug_name = ('inputs', parameter)
                    c_e._remove_plug(
                        _temp_plug_name=temp_plug_name,
                        from_undo=True,
                        from_export_plugs=True)
                self.main_window.statusBar().showMessage(
                    "Plugs {0} have been removed.".format(str(parameter_list)))

            elif action == "remove_plug":
                temp_plug_name = to_undo[1]
                new_temp_plugs = to_undo[2]
                optional = to_undo[3]
                c_e._export_plug(temp_plug_name=new_temp_plugs[0],
                                 weak_link=False,
                                 optional=optional, from_undo=True,
                                 pipeline_parameter=temp_plug_name[1])

                # Connecting all the plugs that were connected
                # to the original plugs
                for plug_tuple in new_temp_plugs:
                    # Checking if the original plug is a pipeline
                    # input or output to adapt
                    # the links to add.
                    if temp_plug_name[0] == 'inputs':
                        source = ('', temp_plug_name[1])
                        dest = plug_tuple
                    else:
                        source = plug_tuple
                        dest = ('', temp_plug_name[1])

                    c_e.scene.add_link(source, dest, active=True, weak=False)

                    # Writing a string to represent the link
                    source_parameters = ".".join(source)
                    dest_parameters = ".".join(dest)
                    link = "->".join((source_parameters, dest_parameters))

                    c_e.scene.pipeline.add_link(link)
                    c_e.scene.update_pipeline()

            elif action == "update_node_name":
                node = to_undo[1]
                new_node_name = to_undo[2]
                old_node_name = to_undo[3]
                c_e.update_node_name(node, new_node_name, old_node_name,
                                     from_undo=True)

            elif action == "update_plug_value":
                node_name = to_undo[1]
                old_value = to_undo[2]
                plug_name = to_undo[3]
                value_type = to_undo[4]
                c_e.update_plug_value(node_name, old_value, plug_name,
                                      value_type, from_undo=True)

            elif action == "add_link":
                link = to_undo[1]
                c_e._del_link(link, from_undo=True)

            elif action == "delete_link":
                source = to_undo[1]
                dest = to_undo[2]
                active = to_undo[3]
                weak = to_undo[4]
                c_e.add_link(source, dest, active, weak, from_undo=True)
                # link = source[0] + "." + source[1] +
                # "->" + dest[0] + "." + dest[1]

            c_e.scene.pipeline.update_nodes_and_plugs_activation()
            self.nodeController.update_parameters()

    def update_clinical_mode(self):
        """
        Update the visibility of widgets/actions depending of the chosen mode

        """
        config = Config()

        # If the clinical mode is chosen, the process library is not available
        # and the user cannot save a pipeline
        if config.get_clinical_mode() == True:
            self.processLibrary.setHidden(True)
            self.previewBlock.setHidden(True)
            self.save_pipeline_action.setDisabled(True)
            self.save_pipeline_as_action.setDisabled(True)
        else:
            self.processLibrary.setHidden(False)
            self.previewBlock.setHidden(False)
            self.save_pipeline_action.setDisabled(False)
            self.save_pipeline_as_action.setDisabled(False)

    def update_project(self, project):
        """
        Update the project attribute of several objects

        :param project: current project in the software
        """
        self.project = project
        self.nodeController.project = project
        self.pipelineEditorTabs.project = project
        self.nodeController.visibles_tags = \
            self.project.session.get_shown_tags()
        self.iterationTable.project = project

        # Necessary for using MIA bricks
        ProcessMIA.project = project

    def update_scans_list(self, iteration_list):
        """
        Update the user-selected list of scans

        :param iteration_list: current list of scans in the iteration table
        """
        if self.iterationTable.check_box_iterate.isChecked():
            self.iteration_table_scans_list = iteration_list
            self.pipelineEditorTabs.scan_list = iteration_list
        else:
            self.iteration_table_scans_list = []
            self.pipelineEditorTabs.scan_list = self.scan_list
        if not self.pipelineEditorTabs.scan_list:
            self.pipelineEditorTabs.scan_list = \
                self.project.session.get_documents_names(COLLECTION_CURRENT)
        self.pipelineEditorTabs.update_scans_list()

    def updateProcessLibrary(self, filename):
        """
        Update the library of processes when a pipeline is saved

        :param filename: file name of the pipeline that has been saved
        """
        filename_folder, file_name = os.path.split(filename)
        module_name = os.path.splitext(file_name)[0]
        class_name = module_name.capitalize()

        tmp_file = os.path.join(filename_folder, module_name + '_tmp')

        # Changing the "Pipeline" class name to the name of file
        with open(filename, 'r') as f:
            with open(tmp_file, 'w') as tmp:
                for line in f:
                    line = line.strip('\r\n')
                    if 'class ' in line:
                        line = 'class {0}(Pipeline):'.format(class_name)
                    tmp.write(line + '\n')

        with open(tmp_file, 'r') as tmp:
            with open(filename, 'w') as f:
                for line in tmp:
                    f.write(line)

        os.remove(tmp_file)
        config = Config()

        if os.path.relpath(filename_folder) != \
                os.path.relpath(os.path.join(
                    config.get_mia_path(), 'processes', 'User_processes')):
            return

        # Updating __init__.py
        init_file = os.path.join(
            config.get_mia_path(), 'processes',
            'User_processes', '__init__.py')

        # Checking that import line is not already in the file
        pattern = 'from .{0} import {1}\n'.format(module_name, class_name)
        file = open(init_file, 'r')
        flines = file.readlines()
        file.close()
        if pattern not in flines:
            with open(init_file, 'a') as f:
                print('from .{0} import {1}'.format(
                    module_name, class_name), file=f)

        package = 'User_processes'
        path = os.path.relpath(os.path.join(filename_folder, '..'))

        # If the pipeline has already been saved
        if 'User_processes.' + module_name in sys.modules.keys():
            # removing the previous version of the module
            del sys.modules['User_processes.' + module_name]
            # this adds the new module version to the sys.modules dictionary
            __import__('User_processes')

        # Adding the module path to the system path
        if path not in sys.path:
            sys.path.insert(0, path)

        self.processLibrary.pkg_library.add_package(package, class_name,
                                                    init_package_tree=True)

        if os.path.relpath(path) not in self.processLibrary.pkg_library.paths:
            self.processLibrary.pkg_library.paths.append(os.path.relpath(path))

        self.processLibrary.pkg_library.save()


class RunProgress(QProgressDialog):
    """Create the pipeline progress bar and launch the thread.

    The progress bar is closed when the thread finishes.

    :param diagram_view: A pipelineEditorTabs
    :param settings: dictionary of settings when the pipeline is iterated
    """

    def __init__(self, diagram_view, settings=None):

        super(RunProgress, self).__init__("Please wait while the pipeline is "
                                          "running...", None, 0, 0)

        if settings:
            self.setWindowTitle(
                "Pipeline is running ({0}/{1})".format(
                    settings["counter"], settings["size"]))
            self.setLabelText('Pipeline is running for {0} in {1}. '
                              'Please wait.'.format(settings['tag_value'],
                                                    settings['tag']))
        else:
            self.setWindowTitle("Pipeline running")
        self.setWindowFlags(
            Qt.Window | Qt.WindowTitleHint | Qt.CustomizeWindowHint)
        self.setModal(True)

        self.diagramView = diagram_view

        self.setMinimumDuration(0)
        self.setValue(0)
        self.setMinimumWidth(350) # For mac OS
        
        self.worker = RunWorker(self.diagramView)
        self.worker.finished.connect(self.close)
        self.worker.start()


class RunWorker(QThread):
    """Run the pipeline"""

    def __init__(self, diagram_view):
        super().__init__()
        self.diagramView = diagram_view

    def run(self):

        def _check_nipype_processes(pplne):
            for node_name, node in pplne.nodes.items():
                if isinstance(node.process, Pipeline):
                    if node_name != "":
                        _check_nipype_processes(node.process)
                elif isinstance(node.process, NipypeProcess):
                    node.process.activate_copy = False

        _check_nipype_processes(self.diagramView.get_current_pipeline())

        pipeline = get_process_instance(
            self.diagramView.get_current_pipeline())

        # Reading config
        config = Config()
        spm_standalone_path = config.get_spm_standalone_path()
        spm_path = config.get_spm_path()
        matlab_path = config.get_matlab_path()
        matlab_standalone_path = config.get_matlab_standalone_path()
        use_spm = config.get_use_spm()
        use_spm_standalone = config.get_use_spm_standalone()
        use_matlab = config.get_use_matlab()

        if use_spm_standalone == True and os.path.exists(
                spm_standalone_path) and os.path.exists(
            matlab_standalone_path):

            if os.path.exists(matlab_path):
                study_config = StudyConfig(
                    use_spm=True, spm_directory=spm_standalone_path,
                    spm_exec=matlab_standalone_path,
                    matlab_exec=matlab_path,
                    output_directory=spm_standalone_path, spm_standalone=True)
            else:
                study_config = StudyConfig(
                    use_spm=True, spm_directory=spm_standalone_path,
                    spm_exec=matlab_standalone_path,
                    output_directory=spm_standalone_path, spm_standalone=True)

        # Using without SPM standalone
        elif use_spm == True and use_matlab == True:
            study_config = StudyConfig(
                use_spm=True, matlab_exec=matlab_path,
                spm_directory=spm_path, spm_standalone=False,
                use_matlab=True, output_directory=spm_path)
        else:
            study_config = StudyConfig(use_spm=False)

        study_config.reset_process_counter()

        try:
            study_config.run(pipeline, verbose=1)
        except OSError as e:
            self.msg = QMessageBox()
            self.msg.setIcon(QMessageBox.Critical)
            self.msg.setText("SPM standalone is not set")
            self.msg.setInformativeText(
                "SPM processes cannot be run with SPM standalone not set.\n"
                "You can activate it and set the paths in MIA preferences.")
            self.msg.setWindowTitle("Error")
            self.msg.setStandardButtons(QMessageBox.Ok)
            self.msg.buttonClicked.connect(self.msg.close)
            self.msg.show()



# -*- coding: utf-8 -*- #
"""
Module to define the advanced search.

Contains:
    Class:
        - AdvancedSearch
"""


##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os

# PyQt5 imports
from PyQt5.QtCore import QObjectCleanupHandler
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import (
    QWidget, QGridLayout, QComboBox, QLineEdit, QPushButton, QHBoxLayout,
    QVBoxLayout, QMessageBox)

# Populse_MIA imports
from populse_mia.utils.tools import ClickableLabel
from populse_mia.software_properties import Config
from populse_mia.data_manager.project import TAG_FILENAME, COLLECTION_CURRENT

# Populse_db imports
from populse_db.database import (
    LIST_TYPES, FIELD_TYPE_STRING, FIELD_TYPE_BOOLEAN)


class AdvancedSearch(QWidget):
    """Class that manages the widget of the advanced search

    The advanced search creates a complex query to the database and is a
    combination of several "query lines" which are linked with AND or OR and
    all composed of:
    - A negation or not
    - A tag name or all visible tags
    - A condition (==, !=, >, <, >=, <=, CONTAINS, IN, BETWEEN)
    - A value

    Methods:
        - add_search_bar: create and define the advanced research bar
        - apply_filter: apply an opened filter to update the table.
        - displayConditionRules: set the list of condition choices,
           depending on the tag type
        - displayValueRules: update the placeholder text when the
           condition choice is changed,
        - get_filters: get the filters in list form
        - launch_search: start the search and update the table
        - prepare_filters: prepare the str representation of the filter
        - refresh_search: refresh the widget
        - remove_row: remove a row
        - rows_borders_added: add the links and the added row to the good rows
        - rows_borders_removed: link and adds row removed from every row
        - show_search: reset the rows when the Advanced Search button is
           clicked,



       - clearLayout: called to clear a layout (not using it resolve issue #72)
       - rowsContainsWidget: check if the widget is still used

    """

    def __init__(self, project, data_browser, scans_list=None,
                 tags_list=None, from_pipeline=False):
        """Initialization of the AdvancedSearch class

        :param project: current project in the software
        :param data_browser: parent data browser widget
        :param scans_list: current list of the documents
        :param tags_list: list of the visualized tags
        :param from_pipeline: True if the widget is called from the pipeline
          manager

        """
        super().__init__()

        if scans_list is None:
            scans_list = []

        if tags_list is None:
            tags_list = []

        self.project = project
        self.dataBrowser = data_browser
        self.rows = []
        self.scans_list = scans_list
        self.tags_list = tags_list
        self.from_pipeline = from_pipeline

        self.search = QPushButton("Search")
        self.search.setFixedWidth(100)

    def add_search_bar(self):
        """Create and define the advanced research bar."""
        row_layout = []

        # NOT choice
        not_choice = QComboBox()
        not_choice.setObjectName('not')
        not_choice.addItem("")
        not_choice.addItem("NOT")

        # Field choice
        field_choice = QComboBox()
        field_choice.setObjectName('field')
        if len(self.tags_list) > 0:
            for tag in self.tags_list:
                field_choice.addItem(tag)
        else:
            for tag in self.project.session.get_shown_tags():
                field_choice.addItem(tag)
        field_choice.model().sort(0)
        field_choice.addItem("All visualized tags")

        # Value choice
        condition_value = QLineEdit()
        condition_value.setObjectName('value')

        # Condition choice
        condition_choice = QComboBox()
        condition_choice.setObjectName('condition')
        condition_choice.addItem("==")
        condition_choice.addItem("!=")
        condition_choice.addItem(">=")
        condition_choice.addItem("<=")
        condition_choice.addItem(">")
        condition_choice.addItem("<")
        condition_choice.addItem("BETWEEN")
        condition_choice.addItem("IN")
        condition_choice.addItem("CONTAINS")
        condition_choice.addItem("HAS VALUE")
        condition_choice.addItem("HAS NO VALUE")
        condition_choice.model().sort(0)

        # Signal to update the placeholder text of the value
        condition_choice.currentTextChanged.connect(
            lambda: self.displayValueRules(condition_choice, condition_value))

        # Signal to update the list of conditions, depending on the tag type
        field_choice.currentTextChanged.connect(
            lambda: self.displayConditionRules(field_choice, condition_choice))

        # Minus to remove the row
        sources_images_dir = Config().getSourceImageDir()
        remove_row_label = ClickableLabel()
        remove_row_picture = QPixmap(os.path.relpath(
            os.path.join(sources_images_dir, "red_minus.png")))
        remove_row_picture = remove_row_picture.scaledToHeight(30)
        remove_row_label.setPixmap(remove_row_picture)

        # Everything appended to the row
        row_layout.append(None)  # Link room
        row_layout.append(not_choice)
        row_layout.append(field_choice)
        row_layout.append(condition_choice)
        row_layout.append(condition_value)
        row_layout.append(remove_row_label)
        row_layout.append(None)  # Add row room

        # Signal to remove the row
        remove_row_label.clicked.connect(lambda: self.remove_row(row_layout))

        self.rows.append(row_layout)

        self.refresh_search()

        self.displayConditionRules(field_choice, condition_choice)

    def apply_filter(self, filter):
        """Apply an opened filter to update the table.

        :param filter: Filter object opened to apply
        """
        self.rows = []

        # Data
        nots = filter.nots
        values = filter.values
        conditions = filter.conditions
        links = filter.links
        fields = filter.fields

        for i in range(0, len(nots)):
            self.add_search_bar()
            row = self.rows[i]
            if i > 0:
                row[0].setCurrentText(links[i - 1])
            row[1].setCurrentText(nots[i])
            row[2].setCurrentText(fields[i][0])

            # Replacing all visualized tags by the current list of visible tags
            if fields[i][0] == "All visualized tags":
                fields[i] = self.project.session.get_shown_tags()

            row[3].setCurrentText(conditions[i])
            row[4].setText(str(values[i]))

        old_rows = self.dataBrowser.table_data.scans_to_visualize

        # Filter applied only if at least one row
        if len(nots) > 0:
            # Result gotten
            try:

                filter_query = self.prepare_filters(
                    links, fields, conditions, values, nots, self.scans_list)
                result = self.project.session.filter_documents(
                    COLLECTION_CURRENT, filter_query)

                # data_browser updated with the new selection
                result_names = [getattr(
                    document, TAG_FILENAME) for document in result]

            except Exception as e:
                print(e)

                # Error message if the search can't be done,
                # and visualization of all scans in the data_browser
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Warning)
                msg.setText(
                    "Error in the search")
                msg.setInformativeText(
                    "The search has encountered a problem, you can correct it "
                    "and launch it again.")
                msg.setWindowTitle("Warning")
                msg.setStandardButtons(QMessageBox.Ok)
                msg.buttonClicked.connect(msg.close)
                msg.exec()
                result_names = self.scans_list

            # data_browser updated with the new selection
            self.dataBrowser.table_data.scans_to_visualize = result_names

        # Otherwise, all the scans are reput
        else:
            # data_browser updated with every scan
            if self.scans_list:
                self.dataBrowser.table_data.scans_to_visualize = \
                    self.scans_list
            else:
                self.dataBrowser.table_data.scans_to_visualize = \
                    self.project.session.get_documents_names(
                        COLLECTION_CURRENT)

        self.dataBrowser.table_data.update_visualized_rows(old_rows)

    # def clearLayout(self, layout):
    #     """
    #     Called to clear a layout
    #
    #     :param layout: layout to clear
    #     """
    #     if layout is not None:
    #         while layout.count():
    #             item = layout.takeAt(0)
    #             widget = item.widget()
    #             # We clear the widget only if the row does not exist anymore
    #             if widget is not None and not self.rowsContainsWidget(
    #                     widget):
    #                 pass
    #                 widget.deleteLater()
    #             else:
    #                 self.clearLayout(item.layout())

    def displayConditionRules(self, field, condition):
        """Set the list of condition choices, depending on the tag type

        :param field: field
        :param condition: condition
        """

        tag_name = field.currentText()
        tag_row = self.project.session.get_field(COLLECTION_CURRENT, tag_name)
        no_operators_tags = []
        for list_type in LIST_TYPES:
            no_operators_tags.append(list_type)
        no_operators_tags.append(FIELD_TYPE_STRING)
        no_operators_tags.append(FIELD_TYPE_BOOLEAN)
        no_operators_tags.append(None)

        if (tag_row is not None and tag_row.type in no_operators_tags) or \
                tag_name == "All visualized tags":
            condition.removeItem(condition.findText("<"))
            condition.removeItem(condition.findText(">"))
            condition.removeItem(condition.findText("<="))
            condition.removeItem(condition.findText(">="))
            condition.removeItem(condition.findText("BETWEEN"))
        elif tag_row is None or tag_row.type not in no_operators_tags:
            operators_to_reput = ["<", ">", "<=", ">=", "BETWEEN"]
            for operator in operators_to_reput:
                is_op_existing = condition.findText(operator) != -1
                if not is_op_existing:
                    condition.addItem(operator)

        if tag_row is not None and tag_row.type in LIST_TYPES:
            condition.removeItem(condition.findText("IN"))
        elif tag_row is None or tag_row.type not in LIST_TYPES:
            operators_to_reput = ["IN"]
            for operator in operators_to_reput:
                is_op_existing = condition.findText(operator) != -1
                if not is_op_existing:
                    condition.addItem(operator)

        condition.model().sort(0)

    def displayValueRules(self, choice, value):
        """Update the placeholder text. when the condition choice is
           changed.

        :param choice: choice
        :param value: value
        """
        if choice.currentText() == "BETWEEN":
            value.setDisabled(False)
            value.setPlaceholderText(
                "Please separate the two inclusive borders of the range by a "
                "semicolon and a space")
        elif choice.currentText() == "IN":
            value.setDisabled(False)
            value.setPlaceholderText("Please separate each list item by a "
                                     "semicolon and a space")
        elif choice.currentText() == "HAS VALUE" \
                or choice.currentText() == "HAS NO VALUE":
            value.setDisabled(True)
            value.setPlaceholderText("")
            value.setText("")
        else:
            value.setDisabled(False)
            value.setPlaceholderText("")

    def get_filters(self, replace_all_by_fields):
        """Get the filters in a list

        :param replace_all_by_fields: to replace All visualized tags by the
           list of visible fields
        :return: Lists of filters (fields, conditions, values, links, nots)
        """

        # Lists to get all the data of the search
        fields = []
        conditions = []
        values = []
        links = []
        nots = []
        for row in self.rows:
            for widget in row:
                if widget is not None:
                    child = widget
                    child_name = child.objectName()
                    if child_name == 'link':
                        links.append(child.currentText())
                    elif child_name == 'condition':
                        conditions.append(child.currentText())
                    elif child_name == 'field':
                        if child.currentText() != "All visualized tags":
                            fields.append([child.currentText()])
                        else:
                            if replace_all_by_fields:
                                fields.append(
                                    self.project.session.get_shown_tags())
                            else:
                                fields.append([child.currentText()])
                    elif child_name == 'value':
                        values.append(child.displayText())
                    elif child_name == 'not':
                        nots.append(child.currentText())

        operators = ["<", ">", "<=", ">=", "BETWEEN"]
        no_operators_tags = []
        for list_type in LIST_TYPES:
            no_operators_tags.append(list_type)
        no_operators_tags.append(FIELD_TYPE_STRING)
        no_operators_tags.append(FIELD_TYPE_BOOLEAN)

        # Converting BETWEEN and IN values into lists
        for i in range(0, len(conditions)):
            if conditions[i] == "BETWEEN" or conditions[i] == "IN":
                values[i] = values[i].split("; ")
            if conditions[i] == "IN":
                for tag in fields[i].copy():
                    tag_row = self.project.session.get_field(
                        COLLECTION_CURRENT, tag)
                    if tag_row.type in LIST_TYPES:
                        fields[i].remove(tag)
            elif conditions[i] in operators:
                for tag in fields[i].copy():
                    tag_row = self.project.session.get_field(
                        COLLECTION_CURRENT, tag)
                    if tag_row.type in no_operators_tags:
                        fields[i].remove(tag)

        return fields, conditions, values, links, nots

    def launch_search(self):
        """Start the search and update the table"""

        # Filters gotten
        (fields, conditions, values, links, nots) = self.get_filters(True)

        old_scans_list = self.dataBrowser.table_data.scans_to_visualize

        try:
            # Result gotten
            filter_query = self.prepare_filters(
                links, fields, conditions, values, nots, self.scans_list)
            result = self.project.session.filter_documents(
                COLLECTION_CURRENT, filter_query)

            # data_browser updated with the new selection
            result_names = [getattr(
                document, TAG_FILENAME) for document in result]

            if not self.from_pipeline:
                self.project.currentFilter.nots = nots
                self.project.currentFilter.values = values
                self.project.currentFilter.fields = fields
                self.project.currentFilter.links = links
                self.project.currentFilter.conditions = conditions

        except Exception as e:
            print(e)

            # Error message if the search can't be done, and visualization
            # of all scans in the databrowser
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Warning)
            msg.setText(
                "Error in the search")
            msg.setInformativeText(
                "The search has encountered a problem, you can correct it "
                "and launch it again.")
            msg.setWindowTitle("Warning")
            msg.setStandardButtons(QMessageBox.Ok)
            msg.buttonClicked.connect(msg.close)
            msg.exec()
            result_names = self.scans_list

        self.dataBrowser.table_data.scans_to_visualize = result_names
        self.dataBrowser.table_data.scans_to_search = result_names
        self.dataBrowser.table_data.update_visualized_rows(old_scans_list)

    @staticmethod
    def prepare_filters(links, fields, conditions, values, nots, scans):
        """Prepare the str representation of the filter

        :param links: list of links (AND/OR)
        :param fields: list of fields
        :param conditions: list of conditions (==, !=, <, >, <=, >=, IN,
           BETWEEN, CONTAINS, HAS VALUE, HAS NO VALUE)
        :param values: list of values
        :param nots: list of negations ("" or NOT)
        :param scans: list of scans to search in
        :return: str representation of the filter
        """

        row_queries = []
        final_query = ""

        # For each row of constraint
        for row in range(0, len(fields)):
            row_fields = fields[row]
            row_condition = conditions[row]
            row_value = values[row]
            row_not = nots[row]

            row_query = "("

            or_to_write = False
            for row_field in row_fields:
                if row_condition == "IN":
                    row_field_query = "({" + row_field + "} " + \
                                      row_condition + " " + \
                                      str(row_value).replace("'", "\"") + ")"
                elif row_condition == "BETWEEN":
                    row_field_query = "(({" + row_field + "} >= \"" + \
                                      row_value[0] + "\") AND (" + row_field\
                                      + " <= \"" + row_value[1] + "\"))"
                elif row_condition == "HAS VALUE":
                    row_field_query = "({" + row_field + "} != null)"
                elif row_condition == "HAS NO VALUE":
                    row_field_query = "({" + row_field + "} == null)"
                elif row_condition == "CONTAINS":
                    row_field_query = "({" + row_field + "} LIKE \"%" + \
                                      row_value + "%\")"
                else:
                    row_field_query = "({" + row_field + "} " + \
                                      row_condition + " \"" + row_value + "\")"

                # Putting OR between conditions if several tags to search in
                if or_to_write:
                    row_field_query = " OR " + row_field_query

                or_to_write = True

                row_query += row_field_query

            row_query += ")"
            row_queries.append(row_query)

            # Negation added if needed
            if row_not == "NOT":
                row_queries[row] = "(NOT " + row_queries[row] + ")"

        final_query += row_queries[0]

        # Putting the link between each row
        for row in range(0, len(links)):
            link = links[row]
            final_query += " " + link + " " + row_queries[row + 1]

        # Taking into account the list of scans
        final_query += " AND ({" + TAG_FILENAME + "} IN " + str(
            scans).replace("'", "\"") + ")"

        final_query = "(" + final_query + ")"

        return final_query

    def refresh_search(self):
        """
        Refresh the widget
        """

        # Old values stored
        (fields, conditions, values, links, nots) = self.get_filters(False)

        # We remove the old layout
        # self.clearLayout(self.layout())
        QObjectCleanupHandler().add(self.layout())

        # Links and add rows removed from every row
        self.rows_borders_removed()

        # Links and add rows put back in the good rows
        self.rows_borders_added(links)

        master_layout = QVBoxLayout()
        main_layout = QGridLayout()

        # Everything added to the layout
        for i in range(0, len(self.rows)):
            for j in range(0, 7):
                widget = self.rows[i][j]
                if widget is not None:
                    main_layout.addWidget(widget, i, j)

        # Search button added at the end
        search_layout = QHBoxLayout(None)
        search_layout.setObjectName("search layout")
        self.search.clicked.connect(self.launch_search)

        search_layout.addWidget(self.search)
        search_layout.setParent(None)

        # New layout added
        master_layout.addLayout(main_layout)
        master_layout.addLayout(search_layout)
        self.setLayout(master_layout)

    def remove_row(self, row_layout):
        """Remove a row

        :param row_layout: Row to remove
        """

        # We remove the row only if there is at least 2 rows, because we
        # always must keep at least one
        if len(self.rows) > 1:
            self.rows.remove(row_layout)

        # We refresh the view
        self.refresh_search()

    def rows_borders_added(self, links):
        """Add the links and the added row to the good rows

        :param links: Old links to reput
        """

        # Plus added to the last row
        sources_images_dir = Config().getSourceImageDir()
        add_search_bar_label = ClickableLabel()
        add_search_bar_label.setObjectName('plus')
        add_search_bar_picture = QPixmap(
            os.path.relpath(
                os.path.join(sources_images_dir, "green_plus.png")))
        add_search_bar_picture = add_search_bar_picture.scaledToHeight(20)
        add_search_bar_label.setPixmap(add_search_bar_picture)
        add_search_bar_label.clicked.connect(self.add_search_bar)
        self.rows[len(self.rows) - 1][6] = add_search_bar_label

        # Link added to every row, except the first one
        for i in range(1, len(self.rows)):
            row = self.rows[i]
            link_choice = QComboBox()
            link_choice.setObjectName('link')
            link_choice.addItem("AND")
            link_choice.addItem("OR")
            if len(links) >= i:
                link_choice.setCurrentText(links[i - 1])
            row[0] = link_choice

    def rows_borders_removed(self):
        """Link and add row removed from every row"""

        # We remove all the links and the add rows
        for i in range(0, len(self.rows)):
            # Plus removed from every row
            if self.rows[i][6] is not None:
                self.rows[i][6].setParent(None)
                self.rows[i][6].deleteLater()
                self.rows[i][6] = None
            # Link removed from every row
            if self.rows[i][0] is not None:
                self.rows[i][0].setParent(None)
                self.rows[i][0].deleteLater()
                self.rows[i][0] = None

    # def rowsContainsWidget(self, widget):
    #     """Check if the widget is still used
    #
    #     :param widget: widget to check
    #     :return: True or False
    #     """
    #     for row in self.rows:
    #         if widget in row:
    #             return True
    #     return False

    def show_search(self):
        """Reset the rows when the Advanced Search button is clicked"""
        self.rows = []
        self.add_search_bar()






# -*- coding: utf-8 -*- #
"""Module that contains the class to handle the projects saved in the software.

Contains:
    Class:
    - SavedProjects

"""

##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

import os
import yaml

# Populse_MIA imports
from populse_mia.software_properties import verCmp
from populse_mia.software_properties import Config


class SavedProjects:
    
    """Class that handles all the projects that have been saved in the software.

    Attributes:
        - savedProjects: dictionary of all the saved projects
        - pathsList: list of all the saved projects

    Methods:
        - addSavedProject: adds a new saved project
        - loadSavedProjects: loads the dictionary from the saved_projects.yml
        file
        - saveSavedProjects: saves the dictionary to the saved_projects.yml
        file
        - removeSavedProject: removes a saved project from the config file

    """

    def __init__(self):
        """Initialise the savedProjects attribute from the saved_projects.yml
           file.

        The pathsList attribute is initialised as the value correponding to 
        the "paths" key in the savedProjects dictionary.

        """
        self.savedProjects = self.loadSavedProjects()

        if ((isinstance(self.savedProjects, dict)) and
            ('paths' in self.savedProjects)):
            self.pathsList = self.savedProjects["paths"]
            
            if self.pathsList is None:
                self.pathsList = []

        else:
            self.savedProjects = {"paths" : []}
            self.pathsList = []

    def addSavedProject(self, newPath):
        """Add a new project to save in the savedProjects and pathsList
           attributes.

        Finally, save the savedProjects attribute in the saved_projects.yml
        file.

        :param newPath: new project's path to add

        :returns: the new path's list (pathsList attribute)

        """""
        if self.pathsList:
            
            if newPath not in self.pathsList:
                self.pathsList.insert(0, newPath)
                
            elif newPath != self.pathsList[0]:
                self.pathsList.remove(newPath)
                self.pathsList.insert(0, newPath)
                
        else:
            self.pathsList.insert(0, newPath)
            
        self.savedProjects["paths"] = self.pathsList
        self.saveSavedProjects()
        return self.pathsList

    def loadSavedProjects(self):
        """Load the savedProjects dictionary from the saved_projects.yml file.

        If the saved_projects.yml file is not existing, it is created with the
        "{paths: []}" value and the returned dictionnary is {paths: []}.

        :returns: the dictionary

        """
        config = Config()

        try:
            
            with open(os.path.join(config.get_mia_path(), 'properties',
                               'saved_projects.yml'), 'r') as stream:
                
                try:
                    if verCmp(yaml.__version__, '5.1', 'sup'):
                        return yaml.load(stream, Loader=yaml.FullLoader)
                    else:    
                        return yaml.load(stream)
            
                except yaml.YAMLError as exc:
                    print(exc)

        except FileNotFoundError as exc:
            
            with open(os.path.join(config.get_mia_path(), 'properties',
                                   'saved_projects.yml'), 'w') as stream:
                yaml.dump({'paths' : []}, stream, default_flow_style=False)

                return {'paths' : []}

    def saveSavedProjects(self):
        """Saves the savedProjects dictionary to the saved_projects.yml file.

        """
        config = Config()
        
        with (open(os.path.join(config.get_mia_path(), 'properties',
                                'saved_projects.yml'), 'w', encoding='utf8')
              ) as configfile:
            yaml.dump(self.savedProjects, configfile,
                      default_flow_style=False, allow_unicode=True)

    def removeSavedProject(self, path):
        """Removes a saved project from the saved_projects.yml file.

        :param path: path of the saved project to remove

        """

        if path in self.savedProjects["paths"]:
            self.savedProjects["paths"].remove(path)

        self.saveSavedProjects()




# -*- coding: utf-8 -*- #
"""Module that contains class to override the default behaviour of
populse_db and some of its methods

Contains:
    Class:
    -DatabaseMIA
    -DatabaseSessionMIA

"""


##########################################################################
# Populse_mia - Copyright (C) IRMaGe/CEA, 2018
# Distributed under the terms of the CeCILL license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html
# for details.
##########################################################################

from sqlalchemy import (
    create_engine, MetaData, String, Boolean, Integer,
    Enum, Column, Table)
from sqlalchemy.exc import ArgumentError
from sqlalchemy.schema import CreateTable
from sqlalchemy.orm import mapper

# Populse_db imports
from populse_db.database import (
    Database, FIELD_TABLE, FIELD_TYPE_STRING,
    FIELD_TYPE_DATE, FIELD_TYPE_LIST_STRING, FIELD_TYPE_LIST_FLOAT,
    FIELD_TYPE_LIST_DATETIME, FIELD_TYPE_LIST_TIME, FIELD_TYPE_LIST_DATE,
    FIELD_TYPE_LIST_INTEGER, FIELD_TYPE_DATETIME, FIELD_TYPE_INTEGER,
    FIELD_TYPE_FLOAT, FIELD_TYPE_TIME, FIELD_TYPE_BOOLEAN,
    FIELD_TYPE_LIST_BOOLEAN, FIELD_TYPE_JSON, FIELD_TYPE_LIST_JSON,
    COLLECTION_TABLE, DatabaseSession, ALL_TYPES, LIST_TYPES, TYPE_TO_COLUMN)
from populse_db.filter import QUERY_MIXED

TAG_ORIGIN_BUILTIN = "builtin"
TAG_ORIGIN_USER = "user"

# Tag unit
TAG_UNIT_MS = "ms"
TAG_UNIT_MM = "mm"
TAG_UNIT_DEGREE = "degree"
TAG_UNIT_HZPIXEL = "Hz/pixel"
TAG_UNIT_MHZ = "MHz"

ALL_UNITS = [TAG_UNIT_MS, TAG_UNIT_MM,
             TAG_UNIT_DEGREE, TAG_UNIT_HZPIXEL, TAG_UNIT_MHZ]


class DatabaseMIA(Database):
    """Class overriding the default behavior of populse_db

    Attributes:
        :param string_engine: Path of the new database file

    Methods:
        - __enter__: returns a DatabaseSession instance for using the database
        - __exit__: releases a DatabaseSession previously created by __enter__
        - _Database__create_empty_schema: overrides the method creating the
        empty schema
    """

    def __init__(self, string_engine):

        super().__init__(string_engine, caches=True, list_tables=True,
                         query_type=QUERY_MIXED)

    def __enter__(self):
        """Return a DatabaseSession instance for using the database. This is
        supposed to be called using a "with" statement.

        Example:
            with database as session:
               session.add_document(...)

        Therefore __exit__ must be called to get rid of the session.
        When called recursively, the underlying database session returned
        is the same. The commit/rollback of the session is done only by the
        outermost __enter__/__exit__ pair (i.e. by the outermost with
        statement).

        :returns: the database session
        """
        # Creates the session object
        new_session = self._Database__scoped_session()
        # Checks if it is a brain new session object or if __enter__ already
        # added a DatabaseSession instance to it (meaning recursive call)
        db_session = getattr(new_session, '_populse_db_session', None)
        if db_session is None:
            # No recursive call. Creates a new DatabaseSession
            # and attaches it to the session. Doing this way allows
            # to be thread safe because scoped_session automatically
            # creates a new session per thread. Therefore we also
            # create a new DatabaseSession per thread.
            db_session = DatabaseSessionMIA(self, new_session)
            new_session._populse_db_session = db_session
            # Attach a counter to the session object to count
            # the recursion depth of __enter__ calls
            new_session._populse_db_counter = 1
        else:
            # __enter__ is called recursively. Simply increment
            # the recursive depth counter previously attached to
            # the session object
            new_session._populse_db_counter += 1
        return db_session

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Release a DatabaseSession previously created by __enter__

        If no recursive call of __enter__ was done, the session
        is committed if no error is reported (e.g. exc_type is None)
        otherwise it is rolled back. Nothing is done
        """
        # Get the current session. SqlAlchemy scoped_session returns
        # the same object (per thread) on each call until remove()
        # is called.
        current_session = self._Database__scoped_session()
        # Decrement recursive depth counter
        current_session._populse_db_counter -= 1
        if current_session._populse_db_counter == 0:
            # If there is no recursive call, commit or rollback
            # the session according to the presence of an exception
            if exc_type is None:
                current_session.commit()
            else:
                current_session.rollback()
            # Delete the database session
            del current_session._populse_db_session
            del current_session._populse_db_counter
            self._Database__scoped_session.remove()

    def _Database__create_empty_schema(self, string_engine):
        """Override the method creating the empty schema, in order to add
        columns to field table

        :param string_engine: Path of the new database file
        """

        try:
            if string_engine.startswith('sqlite'):
                engine = create_engine(string_engine, connect_args={
                    'check_same_thread': False})
            else:
                engine = create_engine(string_engine)
        except ArgumentError:
            raise ValueError("The string engine is invalid, please refer to "
                             "the documentation for more "
                             "details on how to write the string engine")
        metadata = MetaData()
        metadata.reflect(bind=engine)
        if FIELD_TABLE in metadata.tables and COLLECTION_TABLE in \
                metadata.tables:
            return engine
        elif len(metadata.tables) > 0:
            return None
        else:
            Table(FIELD_TABLE, metadata,
                  Column("field_name", String, primary_key=True),
                  Column("collection_name", String, primary_key=True),
                  Column(
                      "type", Enum(FIELD_TYPE_STRING, FIELD_TYPE_INTEGER,
                                   FIELD_TYPE_FLOAT, FIELD_TYPE_BOOLEAN,
                                   FIELD_TYPE_DATE, FIELD_TYPE_DATETIME,
                                   FIELD_TYPE_TIME, FIELD_TYPE_LIST_STRING,
                                   FIELD_TYPE_LIST_INTEGER,
                                   FIELD_TYPE_LIST_FLOAT,
                                   FIELD_TYPE_LIST_BOOLEAN,
                                   FIELD_TYPE_LIST_DATE,
                                   FIELD_TYPE_LIST_DATETIME,
                                   FIELD_TYPE_LIST_TIME, FIELD_TYPE_JSON,
                                   FIELD_TYPE_LIST_JSON),
                      nullable=False),
                  Column("description", String, nullable=True),
                  Column("visibility", Boolean, nullable=False),
                  Column("origin", Enum(TAG_ORIGIN_USER,
                                        TAG_ORIGIN_BUILTIN), nullable=False),
                  Column("unit", Enum(TAG_UNIT_MHZ, TAG_UNIT_DEGREE,
                                      TAG_UNIT_HZPIXEL, TAG_UNIT_MM,
                                      TAG_UNIT_MS), nullable=True),
                  Column("default_value", String, nullable=True))

            Table(COLLECTION_TABLE, metadata,
                  Column("collection_name", String, primary_key=True),
                  Column("primary_key", String, nullable=False))

            metadata.create_all(engine)
            return engine

class DatabaseSessionMIA(DatabaseSession):
    """Class overriding the database session of populse_db

    Methods:
        - add_collection: overrides the method adding a collection
        - add_field: adds a field to the database, if it does not already exist
        - add_fields: adds the list of fields
        - get_shown_tags: gives the list of visible tags
        - set_shown_tags: sets the list of visible tags
    """

    def add_collection(self, name, primary_key, visibility, origin, unit,
                       default_value):
        """Override the method adding a collection of populse_db

        :param name: New collection name
        :param primary_key: New collection primary_key column
        :param visibility: Primary key visibility
        :param origin: Primary key origin
        :param unit: Primary key unit
        :param default_value: Primary key default value
        """

        # Checks
        collection_row = self.get_collection(name)
        if collection_row is not None or name in self.table_classes:
            raise ValueError("A collection/table with the name " +
                             str(name) + " already exists")
        if not isinstance(name, str):
            raise ValueError("The collection name must be of type " + str(str)
                             + ", but collection name of type " +
                             str(type(name)) + " given")
        if not isinstance(primary_key, str):
            raise ValueError("The collection primary_key must be of type " +
                             str(str) +
                             ", but collection primary_key of type " +
                             str(type(primary_key)) + " given")

        # Adding the collection row
        collection_row = self.table_classes[COLLECTION_TABLE](
            collection_name=name, primary_key=primary_key)
        self.session.add(collection_row)

        # Creating the collection document table
        pk_name = self.name_to_valid_column_name(primary_key)
        table_name = self.name_to_valid_column_name(name)
        collection_table = Table(
            table_name, self.metadata, Column(
                pk_name, String, primary_key=True))
        collection_query = CreateTable(collection_table)
        self.session.execute(collection_query)

        # Creating the class associated
        collection_dict = {'__tablename__': table_name,
                           '__table__': collection_table}
        collection_class = type(table_name, (self.base,), collection_dict)
        mapper(collection_class, collection_table)
        self.table_classes[table_name] = collection_class

        # Adding the primary_key of the collection as field
        primary_key_field = self.table_classes[FIELD_TABLE](
            field_name=primary_key, collection_name=name,
            type=FIELD_TYPE_STRING,
            description="Primary_key of the document collection " + name,
            visibility=visibility, origin=origin, unit=unit,
            default_value=default_value)
        self.session.add(primary_key_field)

        if self._DatabaseSession__caches:
            self._DatabaseSession__documents[name] = {}
            self._DatabaseSession__fields[name] = {}
            self._DatabaseSession__fields[name][primary_key] = \
                primary_key_field
            self._DatabaseSession__collections[name] = collection_row

        self.session.flush()

    def add_field(self, collection, name, field_type, description,
                  visibility, origin, unit, default_value,
                  index=False, flush=True):
        """Add a field to the database, if it does not already exist

        :param collection: field collection (str)
        :param name: field name (str)
        :param field_type: field type (string, int, float, boolean, date,
        datetime, time, list_string, list_int, list_float, list_boolean,
        list_date, list_datetime, or list_time)
        :param description: field description (str or None)
        :param visibility: Bool to know if the field is visible in the
        databrowser
        :param origin: To know the origin of a field,
        in [TAG_ORIGIN_BUILTIN, TAG_ORIGIN_USER]
        :param unit: Origin of the field, in [TAG_UNIT_MS, TAG_UNIT_MM,
        TAG_UNIT_DEGREE, TAG_UNIT_HZPIXEL, TAG_UNIT_MHZ]
        :param default_value: Default_value of the field, can be str or None
        :param flush: bool to know if the table classes must be updated (put
        False if in the middle of filling fields) => True by default
        """

        # Checks
        collection_row = self.get_collection(collection)
        if collection_row is None:
            raise ValueError("The collection " +
                             str(collection) + " does not exist")
        field_row = self.get_field(collection, name)
        if field_row is not None:
            raise ValueError("A field with the name " +
                             str(name) + " already exists in the collection " +
                             collection)
        if not isinstance(name, str):
            raise ValueError("The field name must be of type " + str(str) +
                             ", but field name of type " + str(type(name)) +
                             " given")
        if field_type not in ALL_TYPES:
            raise ValueError("The field type must be in " + str(ALL_TYPES) +
                             ", but " + str(field_type) + " given")
        if not isinstance(description, str) and description is not None:
            raise ValueError(
                "The field description must be of type " + str(
                    str) + " or None, but field description of type " + str(
                    type(description)) + " given")

        # Adding the field in the field table
        field_row = self.table_classes[FIELD_TABLE](
            field_name=name, collection_name=collection, type=field_type,
            description=description, visibility=visibility, origin=origin,
            unit=unit, default_value=default_value)

        if self._DatabaseSession__caches:
            self._DatabaseSession__fields[collection][name] = field_row

        self.session.add(field_row)

        # Fields creation
        if field_type in LIST_TYPES:
            if self.list_tables:
                table = 'list_%s_%s' % (self.name_to_valid_column_name(
                    collection), self.name_to_valid_column_name(name))
                list_table = Table(
                    table, self.metadata,
                    Column('document_id', String, primary_key=True),
                    Column('i', Integer, primary_key=True),
                    Column('value', TYPE_TO_COLUMN[field_type[5:]]))
                list_query = CreateTable(list_table)
                self.session.execute(list_query)

                # Creating the class associated
                collection_dict = {'__tablename__': table,
                                   '__table__': list_table}
                collection_class = type(table, (self.base,), collection_dict)
                mapper(collection_class, list_table)
                self.table_classes[table] = collection_class

            # String columns if it list type, as the str representation of the
            # lists will be stored
            field_type = String
        else:
            field_type = self._DatabaseSession__field_type_to_column_type(
                field_type)

        column = Column(
            self.name_to_valid_column_name(name), field_type, index=index)
        column_str_type = column.type.compile(self.database.engine.dialect)
        column_name = column.compile(dialect=self.database.engine.dialect)

        # Column created in document table, and in initial table if initial
        # values are used

        document_query = str('ALTER TABLE "%s" ADD COLUMN %s %s' %
                             (self.name_to_valid_column_name(collection),
                              column_name, column_str_type))
        self.session.execute(document_query)
        self.table_classes[self.name_to_valid_column_name(
            collection)].__table__.append_column(column)

        # Redefinition of the table classes
        if flush:
            self.session.flush()

            # Classes reloaded in order to add the new column attribute
            self._DatabaseSession__update_table_classes()

            if self._DatabaseSession__caches:
                self._DatabaseSession__refresh_cache_documents(collection)

        self.unsaved_modifications = True

    def add_fields(self, fields):
        """Add the list of fields

        :param fields: list of fields (collection, name, type, description,
        visibility, origin, unit, default_value)
        """

        collections = []

        for field in fields:
            # Adding each field
            self.add_field(field[0], field[1], field[2], field[3], field[4],
                           field[5], field[6], field[7], False)
            if field[0] not in collections:
                collections.append(field[0])

        # Updating the table classes
        self.session.flush()

        # Classes reloaded in order to add the new column attribute
        self._DatabaseSession__update_table_classes()

        for collection in collections:
            self._DatabaseSession__refresh_cache_documents(collection)

    def get_shown_tags(self):
        """Give the list of visible tags

        :returns: the list of visible tags
        """

        visible_fields = self.session.query(
            self.table_classes[FIELD_TABLE]).filter(
            self.table_classes[FIELD_TABLE].visibility == True).all()
        visible_names = []
        for field in visible_fields:
            if field.field_name not in visible_names:
                visible_names.append(field.field_name)
        return visible_names

    def set_shown_tags(self, field_showed):
        """Set the list of visible tags

        :param field_showed: list of visible tags
        """

        for field in self.session.query(self.table_classes[FIELD_TABLE]).all():
            if field.field_name in field_showed:
                field.visibility = True
            else:
                field.visibility = False
            self.session.add(field)
        self.session.flush()
        self.unsaved_modifications = True



import os
from setuptools import find_packages, setup

# Select modules to include in distribution
modules = find_packages('python')
print('\nmodules found: ', modules)

# Additional script files to include in distribution
scripts = []

# Data files to include in distribution
pkgdata = {
    # format is:
    # <python module>: <file extensions list>
    # e.g:
    'populse_mia.sources_images': ['*.png', '*.jpeg', '*.jpg']
}

# Read the info.py file in populse_mia module
release_info = {}
python_dir = os.path.join(os.path.dirname(__file__), 'python')
with open(os.path.join(python_dir, 'populse_mia', 'info.py')) as f:
    code = f.read()
    exec(code, release_info)

# Build the setup
setup(
    name=release_info["NAME"],
    description=release_info["DESCRIPTION"],
    long_description=release_info["LONG_DESCRIPTION"],
    license=release_info["LICENSE"],
    classifiers=release_info["CLASSIFIERS"],
    author=release_info["AUTHOR"],
    author_email=release_info["AUTHOR_EMAIL"],
    version=release_info["VERSION"],
    url=release_info["URL"],
    package_dir={'': 'python'},
    packages=modules,
    package_data=pkgdata,
    platforms=release_info["PLATFORMS"],
    install_requires=release_info["REQUIRES"],
    extras_require=release_info["EXTRA_REQUIRES"],
    scripts=scripts
)
